from __future__ import annotations

import concurrent.futures
import logging
import re
import subprocess
import threading
import time
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Callable

logger = logging.getLogger(__name__)


@dataclass
class GrepVerdict:
    alive: bool
    suppression_code: str | None = None
    rationale: str = ""
    evidence: list[str] = field(default_factory=list)


@dataclass
class GrepStrategy:
    name: str
    build_pattern: Callable[..., str | list[str]]
    include_globs: list[str] = field(default_factory=list)
    is_strong: bool = False
    languages: list[str] = field(default_factory=lambda: ["python"])
    use_regex: bool = True
    fixed_string: bool = False
    filter_definitions: bool = True
    result_key: str = ""

    @property
    def key(self) -> str:
        return self.result_key or self.name


_PYTHON_EXTS = {".py", ".pyi"}
_TS_EXTS = {".ts", ".tsx", ".js", ".jsx", ".mjs", ".cjs"}
_GO_EXTS = {".go"}
_JAVA_EXTS = {".java"}
_RUST_EXTS = {".rs"}

_ALL_SOURCE_GLOBS = [
    "*.py",
    "*.pyi",
    "*.ts",
    "*.tsx",
    "*.js",
    "*.jsx",
    "*.mjs",
    "*.cjs",
    "*.go",
    "*.java",
    "*.rs",
    "*.rst",
    "*.md",
    "*.yaml",
    "*.yml",
    "*.toml",
    "*.cfg",
    "*.ini",
    "*.txt",
]

_LANG_GLOBS: dict[str, list[str]] = {
    "python": ["*.py", "*.pyi"],
    "typescript": ["*.ts", "*.tsx", "*.js", "*.jsx", "*.mjs", "*.cjs"],
    "go": ["*.go"],
    "java": ["*.java"],
    "rust": ["*.rs"],
}


def detect_language(file_path: str) -> str:
    ext = Path(file_path).suffix.lower()
    if ext in _PYTHON_EXTS:
        return "python"
    if ext in _TS_EXTS:
        return "typescript"
    if ext in _GO_EXTS:
        return "go"
    if ext in _JAVA_EXTS:
        return "java"
    if ext in _RUST_EXTS:
        return "rust"
    return "python"


def _cached_group_results(
    cache: Any,
    group_name: str,
    finding: dict,
    search_fn: Callable[[], dict[str, list[str]]],
) -> dict[str, list[str]]:
    if cache is None or not group_name:
        return search_fn()

    from skylos.grep_cache import file_content_hash as _fch
    import json as _json

    simple_name = finding.get("simple_name", finding.get("name", ""))
    finding_file = finding.get("file", "")
    content_hash = _fch(finding_file) if finding_file else ""
    cache_key = (
        f"group:{group_name}:{simple_name}:{finding.get('full_name', '')}:"
        f"{finding.get('type', '')}:{content_hash}"
    )
    cached = cache.get(cache_key)
    if cached is not None:
        try:
            return _json.loads(cached[0]) if cached else {}
        except Exception:
            pass

    results = search_fn()
    try:
        cache.put(cache_key, [_json.dumps(results)])
    except Exception:
        pass
    return results


def source_globs_for_language(lang: str) -> list[str]:
    return _LANG_GLOBS.get(lang, _LANG_GLOBS["python"])


def _run_grep(
    pattern: str,
    project_root: str,
    use_regex: bool = False,
    include_globs: list[str] | None = None,
    fixed_string: bool = False,
    max_results: int = 20,
) -> list[str]:
    if include_globs is None:
        include_globs = [
            "*.py",
            "*.rst",
            "*.md",
            "*.yaml",
            "*.yml",
            "*.toml",
            "*.cfg",
            "*.ini",
            "*.txt",
        ]

    grep_flags = ["-rn"]
    if fixed_string:
        grep_flags.append("-F")
    elif use_regex:
        grep_flags.append("-E")

    includes = []
    for g in include_globs:
        includes.extend(["--include", g])

    try:
        cmd = ["grep", *grep_flags, *includes, pattern, project_root]
        result = subprocess.run(cmd, capture_output=True, text=True, timeout=10)
        lines = result.stdout.strip().splitlines()
        return [l for l in lines if "__pycache__" not in l and ".egg-info" not in l][
            :max_results
        ]
    except Exception as e:
        logger.debug("grep failed for pattern %r: %s", pattern, e)
        return []


def repo_relative_path(file_path: str, project_root: str | Path) -> str:
    try:
        rel = Path(file_path).resolve().relative_to(Path(project_root).resolve())
        return rel.as_posix()
    except Exception:
        return Path(file_path).as_posix()


def module_candidates(file_path: str, project_root: str | Path) -> list[str]:
    rel = repo_relative_path(file_path, project_root)
    lang = detect_language(file_path)

    if lang == "python":
        if not rel.endswith(".py"):
            return []
        stem = rel[:-3]
        parts = [p for p in stem.split("/") if p]
        if not parts:
            return []
        if parts[-1] == "__init__":
            parts = parts[:-1]
        if not parts:
            return []
        candidates = [".".join(parts)]
        if parts[0] == "src" and len(parts) > 1:
            candidates.append(".".join(parts[1:]))
        return list(dict.fromkeys(candidates))

    elif lang == "typescript":
        for ext in (".ts", ".tsx", ".js", ".jsx", ".mjs", ".cjs"):
            if rel.endswith(ext):
                stem = rel[: -len(ext)]
                break
        else:
            return []
        parts = [p for p in stem.split("/") if p]
        if not parts:
            return []
        if parts[-1] == "index":
            parts = parts[:-1]
        if not parts:
            return []
        candidates = ["/".join(parts)]
        if parts[0] == "src" and len(parts) > 1:
            candidates.append("/".join(parts[1:]))
        candidates.append(".".join(parts))
        return list(dict.fromkeys(candidates))

    elif lang == "go":
        parts = [p for p in rel.split("/") if p]
        if parts:
            pkg_parts = parts[:-1] if len(parts) > 1 else parts
            return ["/".join(pkg_parts)]
        return []

    elif lang == "java":
        if not rel.endswith(".java"):
            return []
        stem = rel[:-5]
        parts = [p for p in stem.split("/") if p]
        if not parts:
            return []
        for prefix in ("src/main/java", "src/test/java", "src"):
            prefix_parts = prefix.split("/")
            if parts[: len(prefix_parts)] == prefix_parts:
                parts = parts[len(prefix_parts) :]
                break
        return [".".join(parts)] if parts else []

    elif lang == "rust":
        if not rel.endswith(".rs"):
            return []
        stem = rel[:-3]
        parts = [p for p in stem.split("/") if p]
        if not parts:
            return []
        if parts[-1] in ("mod", "lib", "main"):
            parts = parts[:-1]
        if parts and parts[0] == "src":
            parts = parts[1:]
        return ["::".join(parts)] if parts else []

    return []


def parameter_owner_name(finding: dict) -> str:
    if str(finding.get("type", "")).lower() != "parameter":
        return ""
    full_name = str(finding.get("full_name", finding.get("name", "")))
    if "." not in full_name:
        return ""
    return full_name.rsplit(".", 1)[0]


def is_definition_line(grep_line: str, finding: dict) -> bool:
    file_path = finding.get("file", "")
    line_num = finding.get("line", 0)

    if file_path and file_path in grep_line:
        try:
            parts = grep_line.split(":")
            if len(parts) >= 2 and parts[1].strip().isdigit():
                match_line = int(parts[1].strip())
                if abs(match_line - line_num) <= 2:
                    return True
        except (ValueError, IndexError):
            pass

    if ":" in grep_line:
        content = grep_line.split(":", 2)[-1]
    else:
        content = grep_line

    simple_name = finding.get("simple_name", "")
    definition_patterns = [
        # Python
        f"def {simple_name}",
        f"class {simple_name}",
        f"{simple_name} =",
        f'TypeVar("{simple_name}"',
        f"TypeVar('{simple_name}'",
        # TypeScript/JS
        f"function {simple_name}",
        f"const {simple_name}",
        f"let {simple_name}",
        f"var {simple_name}",
        f"interface {simple_name}",
        f"type {simple_name}",
        f"enum {simple_name}",
        f"export default function {simple_name}",
        f"export function {simple_name}",
        f"export const {simple_name}",
        f"export class {simple_name}",
        f"export interface {simple_name}",
        f"export type {simple_name}",
        # Go
        f"func {simple_name}",
        f"type {simple_name} struct",
        f"type {simple_name} interface",
        # Java
        f"public class {simple_name}",
        f"public interface {simple_name}",
        f"private void {simple_name}",
        f"public void {simple_name}",
        f"protected void {simple_name}",
        # Rust
        f"fn {simple_name}",
        f"pub fn {simple_name}",
        f"pub(crate) fn {simple_name}",
        f"struct {simple_name}",
        f"pub struct {simple_name}",
        f"trait {simple_name}",
        f"pub trait {simple_name}",
        f"impl {simple_name}",
    ]
    for pattern in definition_patterns:
        if pattern in content:
            return True

    return False


def filter_grep_results(
    lines: list[str],
    finding: dict,
) -> tuple[list[str], list[str]]:
    """Separate grep results into definitions and usages."""
    definitions = []
    usages = []
    for line in lines:
        if is_definition_line(line, finding):
            definitions.append(line)
        else:
            usages.append(line)
    return definitions, usages


def is_substring_match(grep_line: str, simple_name: str) -> bool:
    """Check if the match is a false positive due to substring matching."""
    if ":" in grep_line:
        content = grep_line.split(":", 2)[-1]
    else:
        content = grep_line

    for match in re.finditer(re.escape(simple_name), content):
        start, end = match.start(), match.end()
        before_ok = start == 0 or not content[start - 1].isalnum()
        after_ok = end == len(content) or not content[end].isalnum()
        if before_ok and after_ok:
            return False
    return True


def _deduplicate_grep_results(
    results: dict[str, list[str]],
) -> dict[str, list[str]]:
    deduped: dict[str, list[str]] = {}

    for strategy, lines in results.items():
        seen_in_strategy: set[str] = set()
        unique = []
        for line in lines:
            parts = line.split(":", 2)
            if len(parts) >= 2 and parts[1].strip().isdigit():
                key = f"{parts[0]}:{parts[1]}"
            else:
                key = line
            if key not in seen_in_strategy:
                seen_in_strategy.add(key)
                unique.append(line)
        if unique:
            deduped[strategy] = unique
        elif strategy in results and not lines:
            deduped[strategy] = lines
    return deduped


_STRONG_ALIVE_STRATEGIES = {
    "references",
    "method_calls",
    "imports",
    "qualified_references",
    "string_dispatch",
    "ts_imports",
    "ts_jsx_usage",
    "go_calls",
    "java_imports",
    "rust_use",
}

_MAX_RESULTS_PER_STRATEGY = 5
_DEFAULT_GREP_WORKERS = 4


def _deterministic_suppress_ts(finding: dict) -> bool:
    file_path = finding.get("file", "")
    simple_name = finding.get("simple_name", finding.get("name", ""))
    kind = finding.get("type", "")

    if any(marker in file_path for marker in (".test.", ".spec.", "__tests__/")):
        if kind in ("function", "class") and simple_name.startswith("test"):
            return True

    if file_path.endswith("index.ts") or file_path.endswith("index.js"):
        if kind == "import":
            return True

    return False


def _deterministic_suppress_go(finding: dict) -> bool:
    simple_name = finding.get("simple_name", finding.get("name", ""))
    file_path = finding.get("file", "")

    if simple_name.startswith("Test") and file_path.endswith("_test.go"):
        return True

    return False


def _deterministic_suppress_java(finding: dict) -> bool:
    decorators = finding.get("decorators", [])
    if isinstance(decorators, list):
        for dec in decorators:
            if str(dec).strip() in (
                "@Test",
                "@Override",
                "@Bean",
                "@Autowired",
                "@Component",
            ):
                return True
    return False


def _deterministic_suppress_rust(finding: dict) -> bool:
    simple_name = finding.get("simple_name", finding.get("name", ""))
    decorators = finding.get("decorators", [])

    if isinstance(decorators, list):
        for dec in decorators:
            dec_str = str(dec).strip()
            if dec_str in ("#[test]", "#[cfg(test)]", "#[derive"):
                return True

    if finding.get("type") == "method":
        full_name = finding.get("full_name", "")
        if "::impl::" in full_name or "::Impl::" in full_name:
            return True

    return False


def _deterministic_suppress_multilang(finding: dict) -> bool:
    lang = detect_language(finding.get("file", ""))
    if lang == "typescript":
        return _deterministic_suppress_ts(finding)
    elif lang == "go":
        return _deterministic_suppress_go(finding)
    elif lang == "java":
        return _deterministic_suppress_java(finding)
    elif lang == "rust":
        return _deterministic_suppress_rust(finding)
    return False


def _run_ts_strategies(
    finding: dict,
    project_root: str,
    max_per_strategy: int,
) -> dict[str, list[str]]:
    simple_name = finding.get("simple_name", finding.get("name", ""))
    kind = finding.get("type", "")
    results: dict[str, list[str]] = {}
    ts_globs = _LANG_GLOBS["typescript"]

    if not simple_name or len(simple_name) <= 1:
        return results

    import_pattern = rf"import\s+.*\b{re.escape(simple_name)}\b"
    import_refs = _run_grep(
        import_pattern,
        project_root,
        use_regex=True,
        include_globs=ts_globs,
        max_results=max_per_strategy,
    )
    if import_refs:
        _defs, usages = filter_grep_results(import_refs, finding)
        if usages:
            results["ts_imports"] = usages[:max_per_strategy]

    # require("...") references
    require_pattern = rf'require\s*\(["\x27].*{re.escape(simple_name)}["\x27]\)'
    require_refs = _run_grep(
        require_pattern,
        project_root,
        use_regex=True,
        include_globs=ts_globs,
        max_results=max_per_strategy,
    )
    if require_refs:
        _defs, usages = filter_grep_results(require_refs, finding)
        if usages:
            results["ts_require"] = usages[:max_per_strategy]

    # JSX usage: <Component /> or <Component>
    if kind in ("class", "function", "variable") and simple_name[0].isupper():
        jsx_pattern = rf"<{re.escape(simple_name)}[\s/>]"
        jsx_refs = _run_grep(
            jsx_pattern,
            project_root,
            use_regex=True,
            include_globs=["*.tsx", "*.jsx"],
            max_results=max_per_strategy,
        )
        if jsx_refs:
            _defs, usages = filter_grep_results(jsx_refs, finding)
            if usages:
                results["ts_jsx_usage"] = usages[:max_per_strategy]

    # Barrel exports: export { X }
    export_pattern = rf"export\s*\{{[^}}]*\b{re.escape(simple_name)}\b"
    export_refs = _run_grep(
        export_pattern,
        project_root,
        use_regex=True,
        include_globs=ts_globs,
        max_results=max_per_strategy,
    )
    if export_refs:
        _defs, usages = filter_grep_results(export_refs, finding)
        if usages:
            results["ts_barrel_export"] = usages[:max_per_strategy]

    # Decorator usage: @Decorator
    if kind in ("class", "function"):
        dec_pattern = rf"@{re.escape(simple_name)}"
        dec_refs = _run_grep(
            dec_pattern,
            project_root,
            use_regex=True,
            include_globs=ts_globs,
            max_results=max_per_strategy,
        )
        if dec_refs:
            _defs, usages = filter_grep_results(dec_refs, finding)
            if usages:
                results["ts_decorator"] = usages[:max_per_strategy]

    # implements Interface
    if kind in ("class", "interface"):
        impl_pattern = rf"implements\s+.*\b{re.escape(simple_name)}\b"
        impl_refs = _run_grep(
            impl_pattern,
            project_root,
            use_regex=True,
            include_globs=ts_globs,
            max_results=max_per_strategy,
        )
        if impl_refs:
            _defs, usages = filter_grep_results(impl_refs, finding)
            if usages:
                results["ts_implements"] = usages[:max_per_strategy]

    return results


def _run_go_strategies(
    finding: dict,
    project_root: str,
    max_per_strategy: int,
) -> dict[str, list[str]]:
    simple_name = finding.get("simple_name", finding.get("name", ""))
    kind = finding.get("type", "")
    results: dict[str, list[str]] = {}
    go_globs = _LANG_GLOBS["go"]

    if not simple_name or len(simple_name) <= 1:
        return results

    call_pattern = rf"\b\w+\.{re.escape(simple_name)}\s*\("
    call_refs = _run_grep(
        call_pattern,
        project_root,
        use_regex=True,
        include_globs=go_globs,
        max_results=max_per_strategy,
    )
    if call_refs:
        _defs, usages = filter_grep_results(call_refs, finding)
        if usages:
            results["go_calls"] = usages[:max_per_strategy]

    if kind == "method":
        iface_pattern = rf"\b{re.escape(simple_name)}\s*\("
        iface_refs = _run_grep(
            iface_pattern,
            project_root,
            use_regex=True,
            include_globs=go_globs,
            max_results=max_per_strategy,
        )
        if iface_refs:
            _defs, usages = filter_grep_results(iface_refs, finding)
            if usages:
                results["go_interface_method"] = usages[:max_per_strategy]

    # Struct field references
    if kind in ("variable", "field"):
        field_pattern = rf"\.{re.escape(simple_name)}\b"
        field_refs = _run_grep(
            field_pattern,
            project_root,
            use_regex=True,
            include_globs=go_globs,
            max_results=max_per_strategy,
        )
        if field_refs:
            _defs, usages = filter_grep_results(field_refs, finding)
            if usages:
                results["go_field_refs"] = usages[:max_per_strategy]

    return results


def _run_java_strategies(
    finding: dict,
    project_root: str,
    max_per_strategy: int,
) -> dict[str, list[str]]:
    simple_name = finding.get("simple_name", finding.get("name", ""))
    kind = finding.get("type", "")
    results: dict[str, list[str]] = {}
    java_globs = _LANG_GLOBS["java"]

    if not simple_name or len(simple_name) <= 1:
        return results

    import_pattern = rf"import\s+.*\b{re.escape(simple_name)}\b"
    import_refs = _run_grep(
        import_pattern,
        project_root,
        use_regex=True,
        include_globs=java_globs,
        max_results=max_per_strategy,
    )
    if import_refs:
        _defs, usages = filter_grep_results(import_refs, finding)
        if usages:
            results["java_imports"] = usages[:max_per_strategy]

    if kind == "method":
        override_pattern = rf"@Override.*\b{re.escape(simple_name)}\b"
        override_refs = _run_grep(
            override_pattern,
            project_root,
            use_regex=True,
            include_globs=java_globs,
            max_results=max_per_strategy,
        )
        if override_refs:
            results["java_override"] = override_refs[:max_per_strategy]

    # implements/extends
    if kind == "class":
        impl_pattern = rf"(?:implements|extends)\s+.*\b{re.escape(simple_name)}\b"
        impl_refs = _run_grep(
            impl_pattern,
            project_root,
            use_regex=True,
            include_globs=java_globs,
            max_results=max_per_strategy,
        )
        if impl_refs:
            _defs, usages = filter_grep_results(impl_refs, finding)
            if usages:
                results["java_implements"] = usages[:max_per_strategy]

    spring_pattern = rf"@\w+.*\b{re.escape(simple_name)}\b"
    spring_refs = _run_grep(
        spring_pattern,
        project_root,
        use_regex=True,
        include_globs=java_globs,
        max_results=max_per_strategy,
    )
    if spring_refs:
        _defs, usages = filter_grep_results(spring_refs, finding)
        if usages:
            results["java_annotations"] = usages[:max_per_strategy]

    return results


def _run_rust_strategies(
    finding: dict,
    project_root: str,
    max_per_strategy: int,
) -> dict[str, list[str]]:
    simple_name = finding.get("simple_name", finding.get("name", ""))
    kind = finding.get("type", "")
    results: dict[str, list[str]] = {}
    rust_globs = _LANG_GLOBS["rust"]

    if not simple_name or len(simple_name) <= 1:
        return results

    use_pattern = rf"use\s+.*\b{re.escape(simple_name)}\b"
    use_refs = _run_grep(
        use_pattern,
        project_root,
        use_regex=True,
        include_globs=rust_globs,
        max_results=max_per_strategy,
    )
    if use_refs:
        _defs, usages = filter_grep_results(use_refs, finding)
        if usages:
            results["rust_use"] = usages[:max_per_strategy]

    if kind in ("class", "trait"):
        impl_pattern = rf"impl\s+.*\b{re.escape(simple_name)}\b"
        impl_refs = _run_grep(
            impl_pattern,
            project_root,
            use_regex=True,
            include_globs=rust_globs,
            max_results=max_per_strategy,
        )
        if impl_refs:
            _defs, usages = filter_grep_results(impl_refs, finding)
            if usages:
                results["rust_impl"] = usages[:max_per_strategy]

    derive_pattern = rf"#\[derive\([^)]*\b{re.escape(simple_name)}\b"
    derive_refs = _run_grep(
        derive_pattern,
        project_root,
        use_regex=True,
        include_globs=rust_globs,
        max_results=max_per_strategy,
    )
    if derive_refs:
        results["rust_derive"] = derive_refs[:max_per_strategy]

    pub_pattern = rf"\b{re.escape(simple_name)}\s*\("
    pub_refs = _run_grep(
        pub_pattern,
        project_root,
        use_regex=True,
        include_globs=rust_globs,
        max_results=max_per_strategy,
    )
    if pub_refs:
        _defs, usages = filter_grep_results(pub_refs, finding)
        if usages:
            results["rust_calls"] = usages[:max_per_strategy]

    return results


def parallel_multi_strategy_search(
    finding: dict,
    project_root: str,
    *,
    max_per_strategy: int = _MAX_RESULTS_PER_STRATEGY,
    early_exit_threshold: int = 5,
    max_workers: int = _DEFAULT_GREP_WORKERS,
    cache: Any = None,
) -> dict[str, list[str]]:
    simple_name = finding.get("simple_name", finding.get("name", ""))
    if not simple_name or len(simple_name) <= 1:
        return {}

    lang = detect_language(finding.get("file", ""))
    results: dict[str, list[str]] = {}
    results_lock = threading.Lock()
    early_exit_event = threading.Event()

    def _check_early_exit() -> bool:
        with results_lock:
            for strategy in _STRONG_ALIVE_STRATEGIES:
                if len(results.get(strategy, [])) >= early_exit_threshold:
                    return True
        return False

    def _run_strategy_group(
        run_fn: Callable[..., dict[str, list[str]]],
        group_name: str = "",
        *args: Any,
    ) -> None:
        if early_exit_event.is_set():
            return
        partial_results = _cached_group_results(
            cache,
            group_name,
            finding,
            lambda: run_fn(*args),
        )
        with results_lock:
            results.update(partial_results)
        if _check_early_exit():
            early_exit_event.set()

    tasks: list[tuple[Callable, tuple]] = []

    if lang == "python":
        tasks.append(
            (
                lambda: multi_strategy_search(
                    finding,
                    project_root,
                    max_per_strategy=max_per_strategy,
                    early_exit_threshold=early_exit_threshold,
                ),
                "python_core",
            )
        )
    else:

        def _general_refs() -> dict[str, list[str]]:
            r: dict[str, list[str]] = {}
            boundary = rf"\b{re.escape(simple_name)}\b"
            refs = _run_grep(
                boundary,
                project_root,
                use_regex=True,
                include_globs=_ALL_SOURCE_GLOBS,
                max_results=max_per_strategy * 2,
            )
            if refs:
                refs = [ref for ref in refs if not is_substring_match(ref, simple_name)]
                _defs, usages = filter_grep_results(refs, finding)
                if usages:
                    r["references"] = usages[:max_per_strategy]
                elif _defs:
                    r["references_definition_only"] = [
                        "(only the definition itself found, no usages)"
                    ]
            return r

        tasks.append((_general_refs, "general_refs"))

    if lang == "typescript":
        tasks.append(
            (
                lambda: _run_ts_strategies(finding, project_root, max_per_strategy),
                "typescript",
            )
        )
    elif lang == "go":
        tasks.append(
            (
                lambda: _run_go_strategies(finding, project_root, max_per_strategy),
                "go",
            )
        )
    elif lang == "java":
        tasks.append(
            (
                lambda: _run_java_strategies(finding, project_root, max_per_strategy),
                "java",
            )
        )
    elif lang == "rust":
        tasks.append(
            (
                lambda: _run_rust_strategies(finding, project_root, max_per_strategy),
                "rust",
            )
        )

    with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = []
        for run_fn, group_name in tasks:
            if early_exit_event.is_set():
                break
            futures.append(executor.submit(_run_strategy_group, run_fn, group_name))

        for future in concurrent.futures.as_completed(futures):
            try:
                future.result(timeout=30)
            except Exception as e:
                logger.debug("Strategy group failed: %s", e)
            if early_exit_event.is_set():
                for f in futures:
                    f.cancel()
                break

    return _deduplicate_grep_results(results)


def multi_strategy_search(
    finding: dict,
    project_root: str,
    *,
    max_per_strategy: int = _MAX_RESULTS_PER_STRATEGY,
    early_exit_threshold: int = 5,
) -> dict[str, list[str]]:
    simple_name = finding.get("simple_name", finding.get("name", ""))
    full_name = finding.get("full_name", "")
    kind = finding.get("type", "")
    file_path = finding.get("file", "")

    if file_path:
        rel_file = repo_relative_path(file_path, project_root)
    else:
        rel_file = ""

    if file_path:
        module_names = module_candidates(file_path, project_root)
    else:
        module_names = []

    if kind == "parameter":
        owner_full_name = parameter_owner_name(finding)
    else:
        owner_full_name = ""

    if owner_full_name:
        owner_simple_name = owner_full_name.rsplit(".", 1)[-1]
    else:
        owner_simple_name = ""

    results: dict[str, list[str]] = {}

    if not simple_name or len(simple_name) <= 1:
        return results

    def _should_early_exit() -> bool:
        for strategy in _STRONG_ALIVE_STRATEGIES:
            hits = results.get(strategy, [])
            if len(hits) >= early_exit_threshold:
                return True
        return False

    boundary_pattern = rf"\b{simple_name}\b"
    refs = _run_grep(
        boundary_pattern, project_root, use_regex=True, max_results=max_per_strategy * 2
    )
    if refs:
        refs = [r for r in refs if not is_substring_match(r, simple_name)]
        _defs, usages = filter_grep_results(refs, finding)
        if usages:
            results["references"] = usages[:max_per_strategy]
        elif _defs:
            results["references_definition_only"] = [
                "(only the definition itself found, no usages)"
            ]

    if _should_early_exit():
        return _deduplicate_grep_results(results)

    if full_name and full_name != simple_name:
        qualified_refs = _run_grep(
            rf"\b{re.escape(full_name)}\b",
            project_root,
            use_regex=True,
            max_results=max_per_strategy,
        )
        if qualified_refs:
            _defs, usages = filter_grep_results(qualified_refs, finding)
            if usages:
                results["qualified_references"] = usages[:max_per_strategy]

    if kind in ("method", "function"):
        call_pattern = rf"\.{simple_name}\s*\("
        call_refs = _run_grep(
            call_pattern,
            project_root,
            use_regex=True,
            include_globs=["*.py"],
            max_results=max_per_strategy,
        )
        if call_refs:
            _defs, usages = filter_grep_results(call_refs, finding)
            if usages:
                results["method_calls"] = usages[:max_per_strategy]

    import_pattern = rf"import.*\b{simple_name}\b"
    import_refs = _run_grep(
        import_pattern,
        project_root,
        use_regex=True,
        include_globs=["*.py"],
        max_results=max_per_strategy,
    )
    if import_refs:
        _defs, usages = filter_grep_results(import_refs, finding)
        if usages:
            results["imports"] = usages[:max_per_strategy]

    if _should_early_exit():
        return _deduplicate_grep_results(results)

    dispatch_patterns = [
        rf'(?:getattr|setattr|hasattr|delattr)\s*\([^,]+,\s*["\x27]{re.escape(simple_name)}["\x27]',
        rf'\[["\x27]{re.escape(simple_name)}["\x27]\]',
        rf'\.\w+\(\s*["\x27]{re.escape(simple_name)}["\x27]',
        rf'["\x27]{re.escape(simple_name)}["\x27]\s*:\s*\w+\(',
    ]
    for dp in dispatch_patterns:
        dp_refs = _run_grep(
            dp,
            project_root,
            use_regex=True,
            include_globs=["*.py"],
            max_results=max_per_strategy,
        )
        if dp_refs:
            dp_refs = [
                r
                for r in dp_refs
                if not any(pat in r for pat in ["TypeVar(", "TypeAlias", "Literal["])
            ]
            _defs, usages = filter_grep_results(dp_refs, finding)
            if usages:
                results["string_dispatch"] = usages[:max_per_strategy]
                break

    if _should_early_exit():
        return _deduplicate_grep_results(results)

    all_refs = _run_grep(
        rf"__all__.*\b{simple_name}\b",
        project_root,
        use_regex=True,
        include_globs=["*.py"],
        max_results=max_per_strategy,
    )
    if all_refs:
        results["exported_in_all"] = all_refs[:max_per_strategy]

    if kind in ("import", "variable", "class"):
        cast_pattern = rf'cast\(\s*["\x27]{simple_name}["\x27]'
        cast_refs = _run_grep(
            cast_pattern,
            project_root,
            use_regex=True,
            include_globs=["*.py"],
            max_results=max_per_strategy,
        )
        if cast_refs:
            _defs, usages = filter_grep_results(cast_refs, finding)
            if usages:
                results["cast_usage"] = usages[:max_per_strategy]

        bound_pattern = rf'bound\s*=\s*["\x27]{simple_name}["\x27]'
        bound_refs = _run_grep(
            bound_pattern,
            project_root,
            use_regex=True,
            include_globs=["*.py"],
            max_results=max_per_strategy,
        )
        if bound_refs:
            _defs, usages = filter_grep_results(bound_refs, finding)
            if usages:
                results["typevar_bound"] = usages[:max_per_strategy]

    elif kind == "method":
        method_parts = full_name.split(".")
        if len(method_parts) >= 2:
            parent_class = method_parts[-2]
            if len(parent_class) > 2:
                cast_pattern = rf"cast\([^,]+,\s*[^)]*\b{parent_class}\b"
                cast_refs = _run_grep(
                    cast_pattern,
                    project_root,
                    use_regex=True,
                    include_globs=["*.py"],
                    max_results=max_per_strategy,
                )
                if cast_refs:
                    _defs, usages = filter_grep_results(cast_refs, finding)
                    if usages:
                        results["cast_protocol"] = usages[:max_per_strategy]

    test_refs = _run_grep(
        rf"\b{simple_name}\b",
        project_root,
        use_regex=True,
        include_globs=["test_*.py", "*_test.py", "conftest.py"],
        max_results=max_per_strategy,
    )
    if test_refs:
        test_refs = [r for r in test_refs if not is_substring_match(r, simple_name)]
        _defs, test_usages = filter_grep_results(test_refs, finding)
        if test_usages:
            results["test_references"] = test_usages[:max_per_strategy]

    if _should_early_exit():
        return _deduplicate_grep_results(results)

    if rel_file and rel_file.endswith(".py"):
        file_refs = _run_grep(
            rel_file, project_root, fixed_string=True, max_results=max_per_strategy
        )
        if file_refs:
            _defs, usages = filter_grep_results(file_refs, finding)
            if usages:
                results["file_path_references"] = usages[:max_per_strategy]

        config_refs = _run_grep(
            rel_file,
            project_root,
            fixed_string=True,
            include_globs=["*.toml", "*.cfg", "*.ini", "*.yaml", "*.yml"],
            max_results=max_per_strategy,
        )
        if config_refs:
            _defs, usages = filter_grep_results(config_refs, finding)
            if usages:
                results["config_references"] = usages[:max_per_strategy]

    for module_name in module_names:
        module_refs = _run_grep(
            module_name, project_root, fixed_string=True, max_results=max_per_strategy
        )
        if module_refs:
            _defs, usages = filter_grep_results(module_refs, finding)
            if usages:
                results["module_references"] = usages[:max_per_strategy]
                break

    if kind == "parameter" and owner_simple_name:
        callback_pattern = (
            rf"callback\s*=\s*(?:[\w\.]+\.)*{re.escape(owner_simple_name)}\b"
        )
        callback_refs = _run_grep(
            callback_pattern,
            project_root,
            use_regex=True,
            include_globs=["*.py"],
            max_results=max_per_strategy,
        )
        if callback_refs:
            results["callback_registrations"] = callback_refs[:max_per_strategy]

        _parse_int = lambda v: int(v) if isinstance(v, (int, float)) else 0
        signature_pattern = rf"def\s+{re.escape(owner_simple_name)}\s*\([^)]*\b{re.escape(simple_name)}\b"
        signature_refs = _run_grep(
            signature_pattern,
            project_root,
            use_regex=True,
            include_globs=["*.py"],
            max_results=max_per_strategy * 2,
        )
        if signature_refs:
            override_refs = []
            line_num = _parse_int(finding.get("line", 0))
            for ref in signature_refs:
                parts = ref.split(":", 2)
                if len(parts) < 2 or not parts[1].isdigit():
                    continue
                match_file = parts[0]
                match_line = int(parts[1])
                if match_file == file_path and abs(match_line - line_num) <= 3:
                    continue
                override_refs.append(ref)
            if override_refs:
                results["signature_overrides"] = override_refs[:max_per_strategy]

    doc_refs = _run_grep(
        rf"\b{simple_name}\b",
        project_root,
        use_regex=True,
        include_globs=["*.rst", "*.md"],
        max_results=max_per_strategy * 2,
    )
    if doc_refs:
        doc_refs = [r for r in doc_refs if not is_substring_match(r, simple_name)]
        if doc_refs:
            compatibility_refs = [
                r
                for r in doc_refs
                if any(
                    keyword in r.lower()
                    for keyword in (
                        "reintroduced",
                        "restored",
                        "backward compatibility",
                        "backwards compatibility",
                        "compatibility",
                        "synonym",
                        "alias",
                        "shim",
                        "shortcut",
                    )
                )
            ]
            if compatibility_refs:
                results["compatibility_references"] = compatibility_refs[
                    :max_per_strategy
                ]
            sphinx_refs = [
                r
                for r in doc_refs
                if any(
                    pat in r
                    for pat in [
                        ":func:",
                        ":meth:",
                        ":class:",
                        ":attr:",
                        "autofunction",
                        "autoclass",
                        "automethod",
                        "automodule",
                        ".. function::",
                        ".. method::",
                    ]
                )
            ]
            if sphinx_refs:
                results["sphinx_directive"] = sphinx_refs[:max_per_strategy]
            else:
                results["doc_references"] = doc_refs[:max_per_strategy]

            if not simple_name.startswith("_"):
                changelog_patterns = [
                    "changelog",
                    "changes",
                    "history",
                    "news",
                    "release",
                ]
                api_refs = []
                for ref in doc_refs:
                    ref_path = ref.split(":", 1)[0].replace("\\", "/").lower()
                    in_docs_dir = (
                        ref_path.startswith("docs/")
                        or ref_path.startswith("doc/")
                        or "/docs/" in ref_path
                        or "/doc/" in ref_path
                    )
                    if not in_docs_dir:
                        continue
                    if any(pattern in ref_path for pattern in changelog_patterns):
                        continue
                    api_refs.append(ref)
                if api_refs:
                    results["public_api_docs"] = api_refs[:max_per_strategy]

    if kind == "method":
        parts = full_name.split(".")
        if len(parts) >= 2:
            class_name = parts[-2]
            if len(class_name) > 2:
                class_refs = _run_grep(
                    rf"\b{class_name}\b",
                    project_root,
                    use_regex=True,
                    include_globs=["*.py"],
                    max_results=max_per_strategy,
                )
                if class_refs:
                    usage_lines = []
                    for cr in class_refs:
                        if ":" in cr:
                            line_text = cr.split(":", 2)[-1]
                        else:
                            line_text = cr
                        if re.search(
                            rf"^\s*class\s+{re.escape(class_name)}", line_text
                        ):
                            continue
                        usage_lines.append(cr)
                    if usage_lines:
                        results["class_usage"] = usage_lines[:max_per_strategy]

    return _deduplicate_grep_results(results)


_DETERMINISTIC_RULES: list[tuple[str, str, str]] = [
    ("method_calls", "real_method_call", "Direct method-call usage found via grep"),
    ("imports", "imported_elsewhere", "Symbol is imported elsewhere in the project"),
    ("string_dispatch", "dynamic_dispatch", "Dynamic dispatch references this symbol"),
    ("qualified_references", "qualified_reference", "Qualified reference found"),
    ("test_references", "test_reference", "Tests reference this symbol"),
    ("config_references", "config_reference", "Referenced in config files"),
    ("cast_protocol", "protocol_required", "Cast to protocol type requires this"),
]


def _apply_deterministic_rules(
    search_results: dict[str, list[str]],
    finding: dict,
) -> GrepVerdict | None:
    refs = search_results.get("references", [])
    if refs:
        simple_name = finding.get("simple_name", "")
        filtered = (
            [r for r in refs if not is_substring_match(r, simple_name)]
            if simple_name
            else refs
        )
        if filtered:
            return GrepVerdict(
                alive=True,
                suppression_code="grep_reference",
                rationale="Grep found usage references in the project",
                evidence=filtered[:3],
            )

    if search_results.get("exported_in_all") and search_results.get("imports"):
        return GrepVerdict(
            alive=True,
            suppression_code="exported_in_all",
            rationale="Exported in __all__ and imported elsewhere",
            evidence=(
                search_results["exported_in_all"][:2] + search_results["imports"][:1]
            ),
        )

    for strategy_key, code, rationale in _DETERMINISTIC_RULES:
        if search_results.get(strategy_key):
            return GrepVerdict(
                alive=True,
                suppression_code=code,
                rationale=rationale,
                evidence=search_results[strategy_key][:3],
            )

    return None


def grep_verify_findings(
    findings: list[dict],
    project_root: str,
    time_budget: float = 30.0,
    *,
    parallel: bool = False,
    max_workers: int = _DEFAULT_GREP_WORKERS,
    cache: Any = None,
) -> dict[str, GrepVerdict]:
    verdicts: dict[str, GrepVerdict] = {}
    start_time = time.monotonic()

    if parallel:

        def search_fn(f: dict) -> dict[str, list[str]]:
            return parallel_multi_strategy_search(
                f, project_root, max_workers=max_workers, cache=cache
            )
    else:

        def search_fn(f: dict) -> dict[str, list[str]]:
            if cache is None:
                return multi_strategy_search(f, project_root)

            lang = detect_language(f.get("file", ""))
            group_name = "python_core" if lang == "python" else f"serial_{lang}"
            return _cached_group_results(
                cache,
                group_name,
                f,
                lambda: multi_strategy_search(f, project_root),
            )

    for finding in findings:
        if time.monotonic() - start_time > time_budget:
            break

        full_name = finding.get("full_name", finding.get("name", ""))
        if not full_name:
            continue

        if _deterministic_suppress_multilang(finding):
            verdicts[full_name] = GrepVerdict(
                alive=True,
                suppression_code="lang_deterministic",
                rationale=f"Language-specific deterministic suppression ({detect_language(finding.get('file', ''))})",
                evidence=[finding.get("file", "")],
            )
            continue

        search_results = search_fn(finding)
        verdict = _apply_deterministic_rules(search_results, finding)
        if verdict:
            verdicts[full_name] = verdict

    return verdicts



import re

from .schemas import Finding, Severity, Confidence, IssueType, CodeLocation


class ValidationResult:
    def __init__(self, valid, adjusted_finding, reason):
        self.valid = valid
        self.adjusted_finding = adjusted_finding
        self.reason = reason


class CodeValidator:
    def __init__(self, strict=False):
        self.strict = strict

    def validate(self, finding, source, file_path):
        lines = source.splitlines()
        total_lines = len(lines)

        issues = []

        location = CodeLocation(
            file=finding.location.file,
            line=finding.location.line,
            end_line=finding.location.end_line,
            column=finding.location.column,
        )

        f = Finding(
            rule_id=finding.rule_id,
            issue_type=finding.issue_type,
            severity=finding.severity,
            confidence=finding.confidence,
            message=finding.message,
            location=location,
            suggestion=finding.suggestion,
            explanation=finding.explanation,
        )

        if f.location.line < 1 or f.location.line > total_lines:
            if self.strict:
                msg = (
                    "Line "
                    + str(f.location.line)
                    + " out of range (1-"
                    + str(total_lines)
                    + ")"
                )
                return ValidationResult(False, None, msg)

            if f.location.line < 1:
                f.location.line = 1
            if f.location.line > total_lines:
                f.location.line = total_lines

            f.confidence = Confidence.LOW
            issues.append("line_adjusted")

        line_content = ""
        if f.location.line >= 1 and f.location.line <= total_lines:
            line_content = lines[f.location.line - 1]

        if not self._is_line_relevant(f, line_content):
            if self.strict:
                return ValidationResult(
                    False, None, "Line content not relevant to finding"
                )

            f.confidence = Confidence.LOW
            issues.append("low_relevance")

        symbol = self._extract_symbol(f.message)
        if symbol is not None and symbol != "":
            if not self._symbol_exists(symbol, source):
                if self.strict:
                    return ValidationResult(
                        False, None, "Symbol '" + symbol + "' not found in code"
                    )

                f.confidence = Confidence.UNCERTAIN
                issues.append("symbol_not_found:" + symbol)

        if f.issue_type == IssueType.SECURITY:
            if not self._verify_security_pattern(f, lines):
                if self.strict:
                    return ValidationResult(
                        False, None, "Security pattern not found at location"
                    )

                f.confidence = Confidence.LOW
                issues.append("pattern_not_verified")

        if issues:
            note = "[Validation: " + ", ".join(issues) + "]"
            if f.explanation:
                f.explanation = str(f.explanation) + " " + note
            else:
                f.explanation = note

        return ValidationResult(True, f, "passed")

    def _is_line_relevant(self, finding, line_content):
        if line_content is None:
            return False

        line = line_content.strip().lower()

        if line == "":
            return False

        if line.startswith("#"):
            return False

        if finding.issue_type == IssueType.SECURITY:
            return True

        if finding.issue_type == IssueType.DEAD_CODE:
            keywords = ["def ", "class ", "import ", "from ", "="]
            for kw in keywords:
                if kw in line:
                    return True
            return False

        if finding.issue_type == IssueType.QUALITY:
            return True

        return True

    def _extract_symbol(self, message):
        if not message:
            return None

        patterns = [
            r"(?:function|method|class|import|variable)[\s:]+['\"`]?(\w+)['\"`]?",
            r"['\"`](\w+)['\"`]",
            r":\s+(\w+)\s*$",
            r"unused\s+(\w+)",
        ]

        for pattern in patterns:
            match = re.search(pattern, message, re.IGNORECASE)
            if match:
                return match.group(1)

        return None

    def _symbol_exists(self, symbol, source):
        if not symbol:
            return False
        if not source:
            return False

        pattern = r"\b" + re.escape(symbol) + r"\b"
        return re.search(pattern, source) is not None

    def _verify_security_pattern(self, finding, lines):
        if finding.location.line > len(lines):
            return False

        start = finding.location.line - 3
        if start < 0:
            start = 0

        end = finding.location.line + 2
        if end > len(lines):
            end = len(lines)

        context = "\n".join(lines[start:end]).lower()
        msg = str(finding.message or "").lower()

        patterns = {
            "sql": [
                "execute",
                "query",
                "cursor",
                "select",
                "insert",
                "update",
                "delete",
                'f"',
                "f'",
                "%s",
                ".format",
            ],
            "injection": ["subprocess", "os.system", "shell", "popen", "eval", "exec"],
            "hardcoded": [
                "password",
                "secret",
                "key",
                "token",
                "api_key",
                "apikey",
                '= "',
                "= '",
            ],
            "xss": ["render", "html", "template", "innerhtml", "mark_safe"],
            "pickle": ["pickle", "loads", "load", "unpickle"],
            "yaml": ["yaml.load", "yaml.unsafe_load"],
            "path": ["../", "path", "file", "open(", "read("],
        }

        for keyword in patterns:
            if keyword in msg:
                indicators = patterns[keyword]
                found = False
                for ind in indicators:
                    if ind in context:
                        found = True
                        break
                if found:
                    return True
                return False

        return True


class ResultValidator:
    def __init__(self, strict=False, min_confidence=Confidence.LOW):
        self.code_validator = CodeValidator(strict=strict)
        self.min_confidence = min_confidence
        self.strict = strict

    def validate(self, findings, source, file_path):
        stats = {
            "original": len(findings),
            "accepted": 0,
            "rejected": 0,
            "adjusted": 0,
        }

        validated = []

        for finding in findings:
            result = self.code_validator.validate(finding, source, file_path)

            if result.valid and result.adjusted_finding is not None:
                if self._meets_confidence(result.adjusted_finding.confidence):
                    validated.append(result.adjusted_finding)
                    stats["accepted"] += 1

                    if result.adjusted_finding.confidence != finding.confidence:
                        stats["adjusted"] += 1
                else:
                    stats["rejected"] += 1
            else:
                stats["rejected"] += 1

        return validated, stats

    def _meets_confidence(self, conf):
        rank = {
            Confidence.HIGH: 0,
            Confidence.MEDIUM: 1,
            Confidence.LOW: 2,
            Confidence.UNCERTAIN: 3,
        }

        conf_rank = rank.get(conf, 999)
        min_rank = rank.get(self.min_confidence, 999)

        return conf_rank <= min_rank


def deduplicate_findings(findings):
    if not findings:
        return []

    seen = set()
    unique = []

    for f in findings:
        msg_key = ""
        if f.message:
            msg_key = f.message[:50].lower()

        is_dup = False

        offset = -2
        while offset <= 2:
            key = (f.location.file, f.location.line + offset, msg_key)
            if key in seen:
                is_dup = True
                break
            offset += 1

        if not is_dup:
            key = (f.location.file, f.location.line, msg_key)
            seen.add(key)
            unique.append(f)

    return unique


def merge_findings(llm_findings, static_findings, file_path):
    merged = []

    llm_by_line = {}
    for f in llm_findings:
        line = f.location.line
        if line not in llm_by_line:
            llm_by_line[line] = []
        llm_by_line[line].append(f)

    static_lines_covered = set()

    for f in llm_findings:
        static_match = False

        for static in static_findings:
            static_line = 0
            if "line" in static:
                static_line = static.get("line", 0)
            else:
                static_line = static.get("lineno", 0)

            if abs(f.location.line - static_line) <= 2:
                static_match = True
                static_lines_covered.add(static_line)
                break

        if static_match:
            if f.confidence == Confidence.MEDIUM or f.confidence == Confidence.LOW:
                f.confidence = Confidence.HIGH

            suffix = " [Corroborated by static analysis]"
            if f.explanation:
                f.explanation = str(f.explanation) + suffix
                f.explanation = f.explanation.strip()
            else:
                f.explanation = suffix.strip()

        merged.append(f)

    for static in static_findings:
        static_line = 0
        if "line" in static:
            static_line = static.get("line", 0)
        else:
            static_line = static.get("lineno", 0)

        if static_line in static_lines_covered:
            continue

        nearby = False

        for line in llm_by_line:
            if abs(line - static_line) <= 2:
                nearby = True
                break

        if nearby:
            continue

        rule_id = static.get("rule_id")
        if not rule_id:
            rule_id = static.get("code", "STATIC")

        msg = static.get("message")
        if not msg:
            msg = static.get("msg", "Issue detected")

        merged.append(
            Finding(
                rule_id=rule_id,
                issue_type=_infer_issue_type(static),
                severity=_parse_severity(static.get("severity", "medium")),
                confidence=Confidence.MEDIUM,
                message=msg,
                location=CodeLocation(file=file_path, line=static_line),
                explanation="[From static analysis only]",
            )
        )

    return merged


def _infer_issue_type(static):
    msg = str(static.get("message", "") + static.get("rule_id", "")).lower()

    sec_words = ["security", "injection", "xss", "sql", "secret"]
    for kw in sec_words:
        if kw in msg:
            return IssueType.SECURITY

    dead_words = ["unused", "dead", "unreachable"]
    for kw in dead_words:
        if kw in msg:
            return IssueType.DEAD_CODE

    return IssueType.QUALITY


def _parse_severity(sev):
    s = str(sev or "").lower()

    if s == "critical" or s == "blocker":
        return Severity.CRITICAL
    if s == "high" or s == "error":
        return Severity.HIGH
    if s == "medium" or s == "warning":
        return Severity.MEDIUM

    return Severity.LOW


Validator = CodeValidator
merge_with_static = merge_findings



from __future__ import annotations

import json
import os
import time
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import Any

from skylos.analyzer import analyze as run_analyze
from skylos.baseline import load_baseline
from skylos.constants import parse_exclude_folders
from skylos.config import load_config
from skylos.file_discovery import discover_source_files

STATE_DIR = ".skylos"
STATE_FILE = "agent_state.json"
SUPPORTED_EXTENSIONS = {".py", ".go", ".ts", ".tsx", ".js", ".jsx"}


def resolve_project_root(path: str | Path) -> Path:
    target = Path(path).resolve()
    if target.is_file():
        target = target.parent

    current = target
    while True:
        if (current / ".git").exists() or (current / "pyproject.toml").exists():
            return current
        parent = current.parent
        if parent == current:
            return target
        current = parent


def resolve_state_path(
    project_root: str | Path, state_file: str | Path | None = None
) -> Path:
    root = Path(project_root).resolve()
    if state_file is None:
        return root / STATE_DIR / STATE_FILE
    path = Path(state_file)
    if not path.is_absolute():
        path = root / path
    return path


def load_agent_state(
    project_root: str | Path, state_file: str | Path | None = None
) -> dict[str, Any] | None:
    path = resolve_state_path(project_root, state_file)
    if not path.exists():
        return None
    try:
        return json.loads(path.read_text(encoding="utf-8"))
    except Exception:
        return None


def save_agent_state(
    project_root: str | Path,
    state: dict[str, Any],
    state_file: str | Path | None = None,
) -> Path:
    path = resolve_state_path(project_root, state_file)
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(json.dumps(state, indent=2) + "\n", encoding="utf-8")
    return path


def snapshot_file_signatures(
    project_root: str | Path,
    *,
    exclude_folders: list[str] | set[str] | None = None,
) -> dict[str, dict[str, int]]:
    root = Path(project_root).resolve()
    excluded = set(
        exclude_folders
        or parse_exclude_folders(
            use_defaults=True,
            config_exclude_folders=load_config(root).get("exclude"),
        )
    )
    excluded.add(STATE_DIR)

    signatures: dict[str, dict[str, int]] = {}
    for path in discover_source_files(root, SUPPORTED_EXTENSIONS, excluded):
        try:
            stat = path.stat()
        except OSError:
            continue
        rel = str(path.relative_to(root)).replace("\\", "/")
        signatures[rel] = {
            "mtime_ns": int(stat.st_mtime_ns),
            "size": int(stat.st_size),
        }
    return signatures


def detect_changed_files(
    previous: dict[str, dict[str, int]] | None,
    current: dict[str, dict[str, int]],
) -> list[str]:
    if not previous:
        return sorted(current.keys())

    changed: set[str] = set()
    for rel_path, signature in current.items():
        if previous.get(rel_path) != signature:
            changed.add(rel_path)
    for rel_path in previous:
        if rel_path not in current:
            changed.add(rel_path)
    return sorted(changed)


def normalize_triage_map(raw: dict[str, Any] | None) -> dict[str, dict[str, Any]]:
    now = datetime.now(timezone.utc)
    triage: dict[str, dict[str, Any]] = {}
    for action_id, entry in (raw or {}).items():
        if not isinstance(entry, dict):
            continue
        status = str(entry.get("status") or "").strip().lower()
        if status not in {"dismissed", "snoozed"}:
            continue
        normalized: dict[str, Any] = {
            "status": status,
            "updated_at": str(entry.get("updated_at") or utc_now()),
        }
        if status == "snoozed":
            snoozed_until = parse_utc_timestamp(entry.get("snoozed_until"))
            if snoozed_until is None or snoozed_until <= now:
                continue
            normalized["snoozed_until"] = snoozed_until.isoformat()
        triage[str(action_id)] = normalized
    return triage


def apply_triage_to_findings(
    findings: list[dict[str, Any]],
    triage: dict[str, dict[str, Any]] | None,
) -> dict[str, int]:
    dismissed = 0
    snoozed = 0
    triage_map = triage or {}
    for finding in findings:
        entry = triage_map.get(finding["fingerprint"]) or {}
        status = entry.get("status")
        finding["triage_status"] = status
        finding["snoozed_until"] = entry.get("snoozed_until")
        finding["is_dismissed"] = status == "dismissed"
        finding["is_snoozed"] = status == "snoozed"
        if finding["is_dismissed"]:
            dismissed += 1
        elif finding["is_snoozed"]:
            snoozed += 1
    return {"dismissed": dismissed, "snoozed": snoozed}


def compose_agent_state(
    project_root: str | Path,
    *,
    signatures: dict[str, dict[str, int]],
    findings: list[dict[str, Any]],
    changed_files: list[str],
    baseline_present: bool,
    triage: dict[str, dict[str, Any]] | None = None,
) -> dict[str, Any]:
    normalized_triage = normalize_triage_map(triage)
    triage_counts = apply_triage_to_findings(findings, normalized_triage)
    actions = build_ranked_actions(findings, changed_files)
    summary = build_summary(
        findings,
        actions,
        changed_files,
        baseline_present,
        triage_counts=triage_counts,
    )

    return {
        "project_root": str(Path(project_root).resolve()),
        "generated_at": utc_now(),
        "state_version": 2,
        "file_signatures": signatures,
        "changed_files": changed_files,
        "baseline_present": baseline_present,
        "triage": normalized_triage,
        "summary": summary,
        "findings": findings,
        "actions": actions,
        "command_center": {
            "headline": summary["headline"],
            "subtitle": summary["subtitle"],
            "items": [
                {
                    "id": action["id"],
                    "title": action["title"],
                    "subtitle": action["subtitle"],
                    "file": action["file"],
                    "absolute_file": action["absolute_file"],
                    "line": action["line"],
                    "severity": action["severity"],
                    "category": action["category"],
                    "score": action["score"],
                    "reason": action["reason"],
                    "action_type": action["action_type"],
                    "command_hint": action["command_hint"],
                    "rule_id": action["rule_id"],
                    "message": action["message"],
                    "safe_fix": action["safe_fix"],
                }
                for action in actions[:10]
            ],
        },
    }


def rebuild_agent_state_from_existing(
    state: dict[str, Any],
    *,
    project_root: str | Path | None = None,
    triage: dict[str, dict[str, Any]] | None = None,
) -> dict[str, Any]:
    root = project_root or state.get("project_root") or "."
    findings: list[dict[str, Any]] = []
    for finding in state.get("findings", []) or []:
        clone = dict(finding)
        clone.pop("triage_status", None)
        clone.pop("snoozed_until", None)
        clone.pop("is_dismissed", None)
        clone.pop("is_snoozed", None)
        findings.append(clone)

    return compose_agent_state(
        root,
        signatures=state.get("file_signatures") or {},
        findings=findings,
        changed_files=list(state.get("changed_files") or []),
        baseline_present=bool(state.get("baseline_present")),
        triage=triage if triage is not None else state.get("triage"),
    )


def update_action_triage(
    path: str | Path,
    action_id: str,
    *,
    status: str,
    state_file: str | Path | None = None,
    snooze_hours: float | None = None,
) -> dict[str, Any]:
    project_root = resolve_project_root(path)
    state = load_agent_state(project_root, state_file=state_file)
    if state is None:
        state, _ = refresh_agent_state(project_root, state_file=state_file, force=True)

    triage = normalize_triage_map(state.get("triage") or {})
    normalized_status = str(status).strip().lower()
    if normalized_status not in {"dismissed", "snoozed"}:
        raise ValueError(f"Unsupported triage status: {status}")

    entry: dict[str, Any] = {
        "status": normalized_status,
        "updated_at": utc_now(),
    }
    if normalized_status == "snoozed":
        if snooze_hours is None or snooze_hours <= 0:
            raise ValueError("snooze_hours must be greater than 0")
        entry["snoozed_until"] = (
            datetime.now(timezone.utc) + timedelta(hours=snooze_hours)
        ).isoformat()

    triage[action_id] = entry
    rebuilt = rebuild_agent_state_from_existing(
        state, project_root=project_root, triage=triage
    )
    save_agent_state(project_root, rebuilt, state_file=state_file)
    return rebuilt


def clear_action_triage(
    path: str | Path,
    action_id: str,
    *,
    state_file: str | Path | None = None,
) -> dict[str, Any]:
    project_root = resolve_project_root(path)
    state = load_agent_state(project_root, state_file=state_file)
    if state is None:
        raise ValueError("No agent state exists yet")

    triage = normalize_triage_map(state.get("triage") or {})
    triage.pop(action_id, None)
    rebuilt = rebuild_agent_state_from_existing(
        state, project_root=project_root, triage=triage
    )
    save_agent_state(project_root, rebuilt, state_file=state_file)
    return rebuilt


def refresh_agent_state(
    path: str | Path,
    *,
    conf: int = 80,
    enable_secrets: bool = True,
    enable_danger: bool = True,
    enable_quality: bool = True,
    include_dead_code: bool = True,
    use_baseline: bool = True,
    state_file: str | Path | None = None,
    force: bool = False,
    exclude_folders: list[str] | set[str] | None = None,
) -> tuple[dict[str, Any], bool]:
    project_root = resolve_project_root(path)
    previous_state = load_agent_state(project_root, state_file=state_file) or {}
    triage = normalize_triage_map(previous_state.get("triage") or {})
    triage_changed = triage != (previous_state.get("triage") or {})
    signatures = snapshot_file_signatures(project_root, exclude_folders=exclude_folders)
    changed_files = detect_changed_files(
        previous_state.get("file_signatures"), signatures
    )

    if not force and previous_state and not changed_files:
        if triage_changed:
            rebuilt = rebuild_agent_state_from_existing(
                previous_state, project_root=project_root, triage=triage
            )
            save_agent_state(project_root, rebuilt, state_file=state_file)
            return rebuilt, True
        return previous_state, False

    raw = run_analyze(
        str(project_root),
        conf=conf,
        enable_secrets=enable_secrets,
        enable_danger=enable_danger,
        enable_quality=enable_quality,
        exclude_folders=list(
            exclude_folders
            or parse_exclude_folders(
                use_defaults=True,
                config_exclude_folders=load_config(project_root).get("exclude"),
            )
        ),
    )
    result = json.loads(raw) if isinstance(raw, str) else raw

    normalized = normalize_findings(
        result, project_root, include_dead_code=include_dead_code
    )
    baseline = load_baseline(project_root) if use_baseline else None
    known = set((baseline or {}).get("fingerprints", []))

    previous_fingerprints = {
        finding.get("fingerprint", "")
        for finding in previous_state.get("findings", [])
        if finding.get("fingerprint")
    }

    for finding in normalized:
        fingerprint = finding["fingerprint"]
        finding["is_new_vs_baseline"] = fingerprint not in known if baseline else True
        finding["is_new_since_last_scan"] = fingerprint not in previous_fingerprints
        finding["is_in_changed_file"] = finding["file"] in changed_files

    state = compose_agent_state(
        project_root,
        signatures=signatures,
        findings=normalized,
        changed_files=changed_files,
        baseline_present=bool(baseline),
        triage=triage,
    )
    save_agent_state(project_root, state, state_file=state_file)
    return state, True


def watch_project(
    path: str | Path,
    *,
    interval: float = 5.0,
    cycles: int | None = None,
    once: bool = False,
    conf: int = 80,
    use_baseline: bool = True,
    state_file: str | Path | None = None,
    exclude_folders: list[str] | set[str] | None = None,
    enable_learning: bool = False,
) -> dict[str, Any]:
    iteration = 0
    latest_state: dict[str, Any] | None = None

    # Initialize grep cache for incremental re-analysis
    grep_cache = None
    try:
        from skylos.grep_cache import GrepCache

        grep_cache = GrepCache()
        project_root = resolve_project_root(path)
        grep_cache.load(str(project_root))
    except ImportError:
        pass

    # Initialize triage learner if enabled
    learner = None
    if enable_learning:
        try:
            from skylos.triage_learner import TriageLearner

            learner = TriageLearner()
            project_root = resolve_project_root(path)
            learner.load(str(project_root))
        except ImportError:
            pass

    previous_fingerprints: set[str] = set()

    while True:
        latest_state, _updated = refresh_agent_state(
            path,
            conf=conf,
            use_baseline=use_baseline,
            state_file=state_file,
            force=iteration == 0 or once,
            exclude_folders=exclude_folders,
        )

        # Track finding lifecycle events
        if latest_state and iteration > 0:
            current_fingerprints = {
                f.get("fingerprint", "")
                for f in latest_state.get("findings", [])
                if f.get("fingerprint")
            }
            appeared = current_fingerprints - previous_fingerprints
            resolved = previous_fingerprints - current_fingerprints
            if appeared:
                latest_state.setdefault("_events", []).append(
                    {
                        "type": "finding_appeared",
                        "count": len(appeared),
                        "iteration": iteration,
                    }
                )
            if resolved:
                latest_state.setdefault("_events", []).append(
                    {
                        "type": "finding_resolved",
                        "count": len(resolved),
                        "iteration": iteration,
                    }
                )
            previous_fingerprints = current_fingerprints
        elif latest_state:
            previous_fingerprints = {
                f.get("fingerprint", "")
                for f in latest_state.get("findings", [])
                if f.get("fingerprint")
            }

        # Save grep cache after each cycle
        if grep_cache and latest_state:
            try:
                grep_cache.save(str(resolve_project_root(path)))
            except Exception:
                pass

        iteration += 1
        if once:
            return latest_state
        if cycles is not None and iteration >= cycles:
            return latest_state
        time.sleep(interval)


def normalize_findings(
    result: dict[str, Any],
    project_root: str | Path,
    *,
    include_dead_code: bool = True,
) -> list[dict[str, Any]]:
    root = Path(project_root).resolve()
    findings: list[dict[str, Any]] = []

    if include_dead_code:
        _append_dead_code(
            findings,
            result.get("unused_functions") or [],
            root,
            "unused_function",
            "INFO",
        )
        _append_dead_code(
            findings, result.get("unused_imports") or [], root, "unused_import", "INFO"
        )
        _append_dead_code(
            findings, result.get("unused_classes") or [], root, "unused_class", "INFO"
        )
        _append_dead_code(
            findings,
            result.get("unused_variables") or [],
            root,
            "unused_variable",
            "INFO",
        )

    _append_findings(findings, result.get("danger") or [], root, "security", "HIGH")
    _append_findings(findings, result.get("secrets") or [], root, "secrets", "HIGH")
    _append_findings(findings, result.get("quality") or [], root, "quality", "MEDIUM")

    findings.sort(
        key=lambda item: (
            -severity_score(item["severity"]),
            item["file"],
            int(item["line"]),
            item["rule_id"],
            item["message"],
        )
    )
    return findings


def build_ranked_actions(
    findings: list[dict[str, Any]], changed_files: list[str]
) -> list[dict[str, Any]]:
    changed = set(changed_files)
    actions: list[dict[str, Any]] = []
    for finding in findings:
        if finding.get("is_dismissed") or finding.get("is_snoozed"):
            continue
        score = severity_score(finding["severity"]) * 100
        if finding.get("is_new_vs_baseline"):
            score += 220
        if finding.get("is_new_since_last_scan"):
            score += 140
        if finding["file"] in changed:
            score += 160
        if finding["category"] in {"security", "secrets"}:
            score += 60
        if finding["category"] == "dead_code":
            score -= 75
        if finding.get("confidence") is not None:
            score += min(int(finding["confidence"]), 100)

        actions.append(
            {
                "id": finding["fingerprint"],
                "title": build_action_title(finding),
                "subtitle": build_action_subtitle(finding),
                "reason": build_action_reason(finding),
                "file": finding["file"],
                "absolute_file": finding["absolute_file"],
                "line": finding["line"],
                "severity": finding["severity"],
                "category": finding["category"],
                "score": score,
                "action_type": infer_action_type(finding),
                "command_hint": f"open:{finding['absolute_file']}:{finding['line']}",
                "rule_id": finding["rule_id"],
                "message": finding["message"],
                "safe_fix": infer_safe_fix(finding),
            }
        )

    actions.sort(
        key=lambda item: (
            -int(item["score"]),
            item["file"],
            int(item["line"]),
            item["title"],
        )
    )
    return actions


def build_summary(
    findings: list[dict[str, Any]],
    actions: list[dict[str, Any]],
    changed_files: list[str],
    baseline_present: bool,
    *,
    triage_counts: dict[str, int] | None = None,
) -> dict[str, Any]:
    critical = sum(
        1 for item in findings if str(item["severity"]).upper() == "CRITICAL"
    )
    high = sum(1 for item in findings if str(item["severity"]).upper() == "HIGH")
    medium = sum(
        1 for item in findings if str(item["severity"]).upper() in {"MEDIUM", "WARN"}
    )
    new_total = sum(1 for item in findings if item.get("is_new_vs_baseline"))
    changed_total = sum(1 for item in findings if item.get("is_in_changed_file"))
    changed_file_count = len(
        {item["file"] for item in findings if item.get("is_in_changed_file")}
    )
    dismissed = int((triage_counts or {}).get("dismissed", 0))
    snoozed = int((triage_counts or {}).get("snoozed", 0))

    headline = build_headline(
        critical=critical,
        high=high,
        new_total=new_total,
        changed_total=changed_total,
        baseline_present=baseline_present,
        total=len(findings),
    )

    subtitle_parts = []
    if changed_file_count:
        subtitle_parts.append(
            f"{changed_total} finding(s) in {changed_file_count} changed file(s)"
        )
    if baseline_present:
        subtitle_parts.append(f"{new_total} new vs baseline")
    if actions:
        subtitle_parts.append(f"{len(actions)} ranked action(s)")
    if snoozed:
        subtitle_parts.append(f"{snoozed} snoozed")
    if dismissed:
        subtitle_parts.append(f"{dismissed} dismissed")
    subtitle = " | ".join(subtitle_parts) if subtitle_parts else "No active actions"

    return {
        "headline": headline,
        "subtitle": subtitle,
        "total_findings": len(findings),
        "new_findings": new_total,
        "critical": critical,
        "high": high,
        "medium": medium,
        "changed_file_count": changed_file_count,
        "changed_files": changed_files,
        "dismissed": dismissed,
        "snoozed": snoozed,
    }


def build_headline(
    *,
    critical: int,
    high: int,
    new_total: int,
    changed_total: int,
    baseline_present: bool,
    total: int,
) -> str:
    urgent = critical + high
    if urgent > 0 and changed_total > 0:
        return f"{urgent} urgent finding(s) need attention in changed code"
    if urgent > 0:
        return f"{urgent} urgent finding(s) need attention"
    if baseline_present and new_total > 0:
        return f"{new_total} new finding(s) since baseline"
    if changed_total > 0:
        return f"{changed_total} finding(s) in files you changed"
    if total > 0:
        return f"{total} tracked finding(s) in repository"
    return "No active findings"


def render_status_table(state: dict[str, Any], *, limit: int = 10) -> dict[str, Any]:
    summary = state.get("summary") or {}
    actions = state.get("actions") or []
    return {
        "headline": summary.get("headline", "No active findings"),
        "subtitle": summary.get("subtitle", ""),
        "actions": actions[:limit],
    }


def command_center_payload(state: dict[str, Any], *, limit: int = 10) -> dict[str, Any]:
    payload = dict(state.get("command_center") or {})
    payload["items"] = list((payload.get("items") or [])[:limit])
    return payload


def _append_findings(
    out: list[dict[str, Any]],
    items: list[dict[str, Any]],
    project_root: Path,
    category: str,
    default_severity: str,
) -> None:
    for item in items:
        file_path = item.get("file", "")
        rel = relative_path(file_path, project_root)
        line = int(item.get("line") or item.get("lineno") or 1)
        rule_id = str(item.get("rule_id") or item.get("rule") or category.upper())
        message = str(item.get("message") or item.get("summary") or rule_id)
        severity = normalize_severity(item.get("severity"), default_severity)
        absolute = str(resolve_file_path(file_path, project_root))
        confidence = item.get("confidence")
        out.append(
            {
                "fingerprint": finding_fingerprint(
                    category, rule_id, rel, line, message
                ),
                "rule_id": rule_id,
                "category": category,
                "severity": severity,
                "message": message,
                "file": rel,
                "absolute_file": absolute,
                "line": line,
                "confidence": confidence,
            }
        )


def _append_dead_code(
    out: list[dict[str, Any]],
    items: list[dict[str, Any]],
    project_root: Path,
    item_type: str,
    severity: str,
) -> None:
    for item in items:
        file_path = item.get("file", "")
        rel = relative_path(file_path, project_root)
        line = int(item.get("line") or item.get("lineno") or 1)
        name = str(item.get("name") or item.get("simple_name") or item_type)
        pretty_type = {
            "unused_function": "function",
            "unused_import": "import",
            "unused_variable": "variable",
            "unused_class": "class",
        }.get(item_type, item_type.replace("_", " "))
        message = f"Unused {pretty_type}: {name}"
        rule_id = dead_code_rule_id(item_type)
        absolute = str(resolve_file_path(file_path, project_root))
        confidence = item.get("confidence")
        out.append(
            {
                "fingerprint": finding_fingerprint(
                    "dead_code", rule_id, rel, line, message
                ),
                "rule_id": rule_id,
                "category": "dead_code",
                "severity": severity,
                "message": message,
                "file": rel,
                "absolute_file": absolute,
                "line": line,
                "confidence": confidence,
            }
        )


def dead_code_rule_id(item_type: str) -> str:
    mapping = {
        "unused_function": "SKY-U001",
        "unused_import": "SKY-U002",
        "unused_variable": "SKY-U003",
        "unused_class": "SKY-U004",
    }
    return mapping.get(item_type, "SKY-U000")


def finding_fingerprint(
    category: str, rule_id: str, file_path: str, line: int, message: str
) -> str:
    return f"{category}:{rule_id}:{file_path}:{line}:{message}"


def relative_path(file_path: str, project_root: Path) -> str:
    try:
        return str(Path(file_path).resolve().relative_to(project_root)).replace(
            "\\", "/"
        )
    except Exception:
        return str(file_path).replace("\\", "/")


def resolve_file_path(file_path: str, project_root: Path) -> Path:
    path = Path(file_path)
    if path.is_absolute():
        return path
    return (project_root / path).resolve()


def normalize_severity(raw: Any, default: str) -> str:
    value = str(raw or default).upper()
    if value == "WARNING":
        return "WARN"
    if value in {"CRITICAL", "HIGH", "MEDIUM", "WARN", "LOW", "INFO"}:
        return value
    return str(default).upper()


def severity_score(severity: str) -> int:
    normalized = str(severity).upper()
    if normalized == "CRITICAL":
        return 5
    if normalized == "HIGH":
        return 4
    if normalized in {"MEDIUM", "WARN"}:
        return 3
    if normalized == "LOW":
        return 2
    return 1


def build_action_title(finding: dict[str, Any]) -> str:
    if finding["category"] == "dead_code":
        return f"Clean up {finding['message']}"
    return f"Review {finding['severity']} {finding['rule_id']}"


def build_action_subtitle(finding: dict[str, Any]) -> str:
    location = f"{finding['file']}:{finding['line']}"
    return f"{finding['message']} ({location})"


def build_action_reason(finding: dict[str, Any]) -> str:
    reasons = []
    if finding.get("is_new_vs_baseline"):
        reasons.append("new vs baseline")
    if finding.get("is_new_since_last_scan"):
        reasons.append("new since last scan")
    if finding.get("is_in_changed_file"):
        reasons.append("in changed file")
    if not reasons:
        reasons.append("ranked by severity")
    return ", ".join(reasons)


def infer_action_type(finding: dict[str, Any]) -> str:
    if finding["category"] == "dead_code":
        return "cleanup"
    if finding["severity"] in {"CRITICAL", "HIGH"}:
        return "inspect_now"
    return "review"


def infer_safe_fix(finding: dict[str, Any]) -> str | None:
    if finding["category"] != "dead_code":
        return None
    if finding["rule_id"] == "SKY-U002":
        return "remove_import"
    if finding["rule_id"] == "SKY-U001":
        return "remove_function"
    return None


def parse_utc_timestamp(value: Any) -> datetime | None:
    if not value:
        return None
    try:
        normalized = str(value)
        if normalized.endswith("Z"):
            normalized = normalized[:-1] + "+00:00"
        parsed = datetime.fromisoformat(normalized)
    except Exception:
        return None
    if parsed.tzinfo is None:
        parsed = parsed.replace(tzinfo=timezone.utc)
    return parsed.astimezone(timezone.utc)


def utc_now() -> str:
    return datetime.now(timezone.utc).isoformat()



from __future__ import annotations

import logging
from collections import defaultdict
from pathlib import Path
from typing import Dict, List, Optional, Set, Tuple

logger = logging.getLogger("Skylos")

CONVENTION_ENTRY_FILES = frozenset(
    {
        "__main__.py",
        "manage.py",
        "wsgi.py",
        "asgi.py",
        "conftest.py",
        "setup.py",
        "fabfile.py",
        "tasks.py",
        "app.py",
        "main.py",
        "server.py",
        "cli.py",
    }
)


class ModuleReachabilityAnalyzer:
    def __init__(self):
        self.graph: Dict[str, Set[str]] = defaultdict(set)
        self.all_modules: Set[str] = set()
        self.entry_points: Set[str] = set()
        self._getattr_packages: Set[str] = set()
        self._dynamic_import_modules: Set[str] = set()

    def build(
        self,
        modmap: Dict[Path, str],
        all_raw_imports: Dict[Path, List[Tuple]],
        pyproject_entrypoints: Optional[Set[str]] = None,
        dynamic_modules: Optional[Set[str]] = None,
        file_defs: Optional[Dict[Path, list]] = None,
    ) -> None:
        for file_path, mod_name in modmap.items():
            if not mod_name:
                continue
            self.all_modules.add(mod_name)

        if not self.all_modules:
            return

        known_roots = set()
        for m in self.all_modules:
            known_roots.add(m.split(".")[0])

        for file_path, raw_imports in all_raw_imports.items():
            from_mod = modmap.get(file_path)
            if not from_mod:
                continue

            for import_module, *_ in raw_imports:
                if not import_module:
                    continue
                root = import_module.split(".")[0]
                if root not in known_roots:
                    continue

                target = self._resolve_target(import_module)
                if target:
                    self.graph[from_mod].add(target)

        self._detect_entry_points(modmap, pyproject_entrypoints)
        self._detect_getattr_packages(modmap, file_defs)
        self._expand_getattr_packages()

        if dynamic_modules:
            self._dynamic_import_modules = set(dynamic_modules)

    def _resolve_target(self, import_path: str) -> Optional[str]:
        parts = import_path.split(".")
        for i in range(len(parts), 0, -1):
            candidate = ".".join(parts[:i])
            if candidate in self.all_modules:
                return candidate
        return None

    def _detect_entry_points(
        self,
        modmap: Dict[Path, str],
        pyproject_entrypoints: Optional[Set[str]] = None,
    ) -> None:
        for file_path, mod_name in modmap.items():
            if not mod_name:
                continue
            basename = Path(file_path).name
            if basename in CONVENTION_ENTRY_FILES:
                self.entry_points.add(mod_name)

        for file_path, mod_name in modmap.items():
            if not mod_name:
                continue
            basename = Path(file_path).name
            if basename.startswith("test_") or basename.endswith("_test.py"):
                self.entry_points.add(mod_name)

        if pyproject_entrypoints:
            for qname in pyproject_entrypoints:
                parts = qname.rsplit(".", 1)
                if len(parts) == 2:
                    mod = parts[0]
                    resolved = self._resolve_target(mod)
                    if resolved:
                        self.entry_points.add(resolved)
                resolved = self._resolve_target(qname)
                if resolved:
                    self.entry_points.add(resolved)

        for file_path, mod_name in modmap.items():
            if not mod_name:
                continue
            parts = mod_name.split(".")
            basename = Path(file_path).name
            if len(parts) == 1 and basename == "__init__.py":
                self.entry_points.add(mod_name)

    def _detect_getattr_packages(
        self,
        modmap: Dict[Path, str],
        file_defs: Optional[Dict[Path, list]] = None,
    ) -> None:
        if not file_defs:
            return

        for file_path, defs in file_defs.items():
            basename = Path(file_path).name
            if basename != "__init__.py":
                continue

            mod_name = modmap.get(file_path)
            if not mod_name:
                continue

            for defn in defs:
                simple = getattr(defn, "simple_name", None)
                def_type = getattr(defn, "type", None)
                if simple == "__getattr__" and def_type == "function":
                    self._getattr_packages.add(mod_name)
                    break

    def _expand_getattr_packages(self) -> None:
        for pkg in self._getattr_packages:
            prefix = pkg + "."
            for mod in self.all_modules:
                if mod.startswith(prefix):
                    self.graph[pkg].add(mod)
        self._refine_getattr_expansion()

    def _refine_getattr_expansion(self) -> None:
        for pkg in self._getattr_packages:
            prefix = pkg + "."
            pkg_depth = pkg.count(".") + 1

            direct_children = set()
            for mod in self.all_modules:
                if mod.startswith(prefix):
                    mod_depth = mod.count(".")
                    if mod_depth == pkg_depth:
                        direct_children.add(mod)

            to_remove = set()
            for mod in list(self.graph.get(pkg, set())):
                if not mod.startswith(prefix):
                    continue
                mod_depth = mod.count(".")
                if mod_depth <= pkg_depth:
                    continue

                has_import_edge = False
                for child in direct_children:
                    if mod in self.graph.get(child, set()):
                        has_import_edge = True
                        break

                if not has_import_edge:
                    to_remove.add(mod)

            for mod in to_remove:
                self.graph[pkg].discard(mod)

    def find_unreachable(self) -> Set[str]:
        if not self.all_modules:
            return set()

        if not self.entry_points:
            logger.debug(
                "Module reachability: no entry points detected, skipping analysis"
            )
            return set()

        reachable: Set[str] = set()
        queue = list(self.entry_points)

        while queue:
            current = queue.pop()
            if current in reachable:
                continue
            reachable.add(current)

            for target in self.graph.get(current, set()):
                if target not in reachable:
                    queue.append(target)

            current_root = current.split(".")[0]
            if current_root in (self._dynamic_import_modules or set()):
                for mod in self.all_modules:
                    if mod.split(".")[0] == current_root and mod not in reachable:
                        queue.append(mod)

            for pkg in self._getattr_packages:
                if current.startswith(pkg + ".") or current == pkg:
                    if pkg not in reachable:
                        queue.append(pkg)

        unreachable = self.all_modules - reachable
        if unreachable:
            logger.info(
                f"Module reachability: {len(unreachable)}/{len(self.all_modules)} "
                f"modules unreachable from entry points"
            )

        return unreachable


def find_unreachable_modules(
    modmap: Dict[Path, str],
    all_raw_imports: Dict[Path, List[Tuple]],
    pyproject_entrypoints: Optional[Set[str]] = None,
    dynamic_modules: Optional[Set[str]] = None,
    file_defs: Optional[Dict[Path, list]] = None,
) -> Set[str]:
    analyzer = ModuleReachabilityAnalyzer()
    analyzer.build(
        modmap=modmap,
        all_raw_imports=all_raw_imports,
        pyproject_entrypoints=pyproject_entrypoints,
        dynamic_modules=dynamic_modules,
        file_defs=file_defs,
    )
    return analyzer.find_unreachable()



from pathlib import Path
from skylos.rules.danger.danger import scan_ctx


def _write(tmp_path: Path, name, code):
    p = tmp_path / name
    p.write_text(code, encoding="utf-8")
    return p


def _rule_ids(findings):
    return {f["rule_id"] for f in findings}


def _scan_one(tmp_path: Path, name, code):
    file_path = _write(tmp_path, name, code)
    return scan_ctx(tmp_path, [file_path])


def test_open_tainted_flags(tmp_path):
    code = (
        "def f(p):\n"
        "    with open(p, 'r', encoding='utf-8', errors='ignore') as fh:\n"
        "        fh.read()\n"
    )
    out = _scan_one(tmp_path, "pt_open.py", code)
    assert "SKY-D215" in _rule_ids(out)


def test_os_remove_tainted_flags(tmp_path):
    code = "import os\ndef f(p):\n    os.remove(p)\n"
    out = _scan_one(tmp_path, "pt_os.py", code)
    assert "SKY-D215" in _rule_ids(out)


def test_open_constant_ok(tmp_path):
    code = "def f():\n    open('README.md', 'r')\n"
    out = _scan_one(tmp_path, "pt_ok.py", code)
    assert "SKY-D215" not in _rule_ids(out)



from skylos.visitors.languages.typescript import scan_typescript_file

TS_CODE = """
// 1. DEAD CODE
function unusedLegacyFunction() {
    return "delete me";
}

// 2. DANGER
function runUnsafe(input: string) {
    eval(input);
}

// 3. QUALITY (Complexity ~3)
function simpleFunction(x: number) {
    if (x > 10) { return true; } else { return false; }
}

runUnsafe("test");
"""


def test_typescript_scanner_defaults(tmp_path):
    d = tmp_path / "src"
    d.mkdir()
    p = d / "app.ts"
    p.write_text(TS_CODE, encoding="utf-8")

    results = scan_typescript_file(str(p))
    defs, refs, _, _, _, _, quality, danger, _, _, _, _, _ = results

    def_names = {d.name for d in defs}
    ref_names = {r[0] for r in refs}

    assert "unusedLegacyFunction" in def_names
    assert "unusedLegacyFunction" not in ref_names
    assert "runUnsafe" in ref_names

    eval_findings = [f for f in danger if f["rule_id"] == "SKY-D201"]
    assert len(eval_findings) == 1

    assert len(quality) == 0


def test_typescript_config_override(tmp_path):
    d = tmp_path / "src"
    d.mkdir()
    p = d / "app.ts"
    p.write_text(TS_CODE, encoding="utf-8")

    strict_config = {"languages": {"typescript": {"complexity": 1}}}

    results = scan_typescript_file(str(p), config=strict_config)
    _, _, _, _, _, _, quality, _, _, _, _, _, _ = results

    assert len(quality) > 0
    assert quality[0]["rule_id"] == "SKY-Q301"



from abc import ABC, abstractmethod


class BaseAdapter(ABC):
    def __init__(self, model, api_key):
        self.model = model
        self.api_key = api_key

    @abstractmethod
    def complete(self, system_prompt, user_prompt, response_format=None):
        pass




import sys


STEPS = [
    {
        "title": "Welcome to Skylos",
        "step": 1,
        "body": (
            "[bold]Skylos[/bold] finds dead code, security vulnerabilities, and quality\n"
            "issues in [bold cyan]Python[/bold cyan], [bold cyan]TypeScript[/bold cyan], and [bold cyan]Go[/bold cyan] codebases.\n"
            "\n"
            "It combines static analysis with optional LLM verification\n"
            "for results that match human-level accuracy."
        ),
    },
    {
        "title": "Quick Scan",
        "step": 2,
        "body": (
            "[bold]Basic scan[/bold] (dead code only):\n"
            "  [bold cyan]skylos .[/bold cyan]\n"
            "\n"
            "[bold]Full scan[/bold] (dead code + security + secrets + quality):\n"
            "  [bold cyan]skylos . -a[/bold cyan]\n"
            "\n"
            "[bold]Scan specific paths[/bold]:\n"
            "  [bold cyan]skylos src/ lib/[/bold cyan]\n"
            "\n"
            "[dim]Output formats: --json, --sarif, --tree, --tui[/dim]"
        ),
    },
    {
        "title": "AI Agent",
        "step": 3,
        "body": (
            "[bold]LLM-verified dead code[/bold] (100% accuracy on benchmarks):\n"
            "  [bold cyan]skylos agent verify .[/bold cyan]\n"
            "\n"
            "[bold]Auto-fix and create PR[/bold]:\n"
            "  [bold cyan]skylos agent remediate .[/bold cyan]\n"
            "\n"
            "[bold]Hybrid static + LLM scan[/bold]:\n"
            "  [bold cyan]skylos agent scan .[/bold cyan]\n"
            "\n"
            "[dim]Requires an API key: skylos key[/dim]"
        ),
    },
    {
        "title": "AI Defense",
        "step": 4,
        "body": (
            "[bold]Map LLM integrations[/bold] in your codebase:\n"
            "  [bold cyan]skylos discover .[/bold cyan]\n"
            "\n"
            "[bold]Check for missing defenses[/bold] (OWASP LLM Top 10):\n"
            "  [bold cyan]skylos defend .[/bold cyan]\n"
            "\n"
            "Detects prompt injection risks, missing PII filters,\n"
            "RAG isolation gaps, rate limiting, and more."
        ),
    },
    {
        "title": "CI/CD Integration",
        "step": 5,
        "body": (
            "[bold]Generate GitHub Actions workflow[/bold] (30-second setup):\n"
            "  [bold cyan]skylos cicd init[/bold cyan]\n"
            "\n"
            "[bold]Quality gate[/bold] (fail CI on findings):\n"
            "  [bold cyan]skylos cicd gate[/bold cyan]\n"
            "\n"
            "[bold]Post inline PR comments[/bold] with LLM-generated fixes:\n"
            "  [bold cyan]skylos cicd review[/bold cyan]"
        ),
    },
    {
        "title": "What's Next",
        "step": 6,
        "body": (
            "[bold]See all commands:[/bold]\n"
            "  [bold cyan]skylos commands[/bold cyan]\n"
            "\n"
            "[bold]Get help on any command:[/bold]\n"
            "  [bold cyan]skylos <command> --help[/bold cyan]\n"
            "\n"
            "[bold]Connect to Skylos Cloud:[/bold]\n"
            "  [bold cyan]skylos login[/bold cyan]\n"
            "\n"
            "[bold]Docs:[/bold] [blue]github.com/duriantaco/skylos[/blue]"
        ),
    },
]

TOTAL_STEPS = len(STEPS)


def run_tour(console):
    from rich.panel import Panel

    is_tty = sys.stdin.isatty()

    for i, step in enumerate(STEPS):
        title = f"[bold cyan][{step['step']}/{TOTAL_STEPS}] {step['title']}[/bold cyan]"

        console.print()
        console.print(Panel(
            step["body"],
            title=title,
            border_style="cyan",
            padding=(1, 2),
        ))

        if is_tty and i < len(STEPS) - 1:
            try:
                input("\n  Press Enter to continue...")
            except (EOFError, KeyboardInterrupt):
                console.print()
                return

    console.print()
    console.print("  [bold green]Tour complete![/bold green] Run [bold]skylos .[/bold] to start scanning.")
    console.print()



from __future__ import annotations
from pathlib import Path
from typing import Optional

from rich.console import Console

console = Console()

ANALYSIS_FLAG_MAP: dict[str, str] = {
    "dead-code": "",
    "security": "--danger",
    "quality": "--quality",
    "secrets": "--secrets",
}


def generate_workflow(
    *,
    triggers: Optional[list[str]] = None,
    analysis_types: Optional[list[str]] = None,
    python_version: str = "3.12",
    use_baseline: bool = True,
    use_llm: bool = False,
    model: Optional[str] = None,
    use_claude_security: bool = False,
    use_upload: bool = False,
    use_defend: bool = False,
) -> str:
    triggers = triggers or ["pull_request", "push"]
    analysis_types = analysis_types or ["dead-code", "security", "quality", "secrets"]

    trigger_block = _build_trigger_block(triggers)
    analysis_flags = " ".join(
        ANALYSIS_FLAG_MAP[t]
        for t in analysis_types
        if t in ANALYSIS_FLAG_MAP and ANALYSIS_FLAG_MAP[t]
    )
    if analysis_flags:
        analysis_flags = " " + analysis_flags

    upload_flag = ""
    upload_env_block = ""
    if use_upload:
        upload_flag = " --upload"
        upload_env_block = (
            "\n        env:\n          SKYLOS_TOKEN: ${{ secrets.SKYLOS_TOKEN }}"
        )

    llm_env = ""
    llm_step = ""
    if use_llm:
        model_str = model or "gpt-4.1"
        api_key_env = ""
        if model_str and "claude" in model_str.lower():
            api_key_env = (
                "\n          ANTHROPIC_API_KEY: ${{ secrets.ANTHROPIC_API_KEY }}"
            )
        llm_step = "\n".join(
            [
                "",
                "      - name: Skylos Agent Review (LLM)",
                "        if: github.event_name == 'pull_request'",
                f"        run: skylos agent review . --model {model_str} --format json -o skylos-llm-results.json",
                "        env:",
                "          SKYLOS_API_KEY: ${{ secrets.SKYLOS_API_KEY }}" + api_key_env,
            ]
        )
        llm_env = """
          SKYLOS_API_KEY: ${{ secrets.SKYLOS_API_KEY }}"""

    defend_step = ""
    if use_defend:
        defend_env = ""
        if use_upload:
            defend_env = (
                "\n        env:\n          SKYLOS_TOKEN: ${{ secrets.SKYLOS_TOKEN }}"
            )
        defend_parts = [
            "",
            "      - name: AI Defense Check",
            f"        run: skylos defend . --fail-on critical --min-score 70 --json -o defense-results.json{' --upload' if use_upload else ''}",
        ]
        if defend_env:
            defend_parts.append(defend_env)
        defend_step = "\n".join(defend_parts)

    permissions_block = """permissions:
  contents: read
  pull-requests: write
  checks: write"""
    if use_claude_security:
        permissions_block += "\n  security-events: write"

    workflow = f"""name: Skylos Analysis

{trigger_block}

{permissions_block}

jobs:
  skylos:
    runs-on: ubuntu-latest
    steps:
      - name: Checkout
        uses: actions/checkout@v4
        with:
          fetch-depth: 0

      - name: Setup Python
        uses: actions/setup-python@v5
        with:
          python-version: '{python_version}'

      - name: Install Skylos
        run: pip install skylos

      - name: Run Skylos Analysis
        run: skylos .{analysis_flags}{upload_flag} --json -o skylos-results.json{
        upload_env_block
    }
{llm_step}{defend_step}
      - name: Quality Gate
        if: always()
        run: skylos cicd gate --input skylos-results.json --summary

      - name: GitHub Annotations
        if: always()
        run: skylos cicd annotate --input skylos-results.json

      - name: PR Review Comments
        if: github.event_name == 'pull_request' && always()
        run: skylos cicd review --input skylos-results.json{
        " --llm-input skylos-llm-results.json" if use_llm else ""
    } --diff-base origin/${{{{ github.base_ref || 'main' }}}}
        env:
          GH_TOKEN: ${{{{ github.token }}}}
{
        ""
        if not use_claude_security
        else '''
      - name: Upload Skylos Results for Cross-Reference
        if: always()
        uses: actions/upload-artifact@v4
        with:
          name: skylos-results
          path: skylos-results.json
'''
    }"""

    if use_claude_security:
        workflow += _build_claude_security_jobs(python_version)

    return workflow


def _build_claude_security_jobs(python_version: str) -> str:
    return f"""
  claude-security:
    runs-on: ubuntu-latest
    steps:
      - name: Checkout
        uses: actions/checkout@v4
        with:
          fetch-depth: 0

      - name: Run Claude Code Security Review
        uses: anthropics/claude-code-action@main
        with:
          anthropic_api_key: ${{{{ secrets.ANTHROPIC_API_KEY }}}}
          direct_prompt: "/review --output-file claude-security-results.json"

      - name: Upload Claude Security Results
        uses: actions/upload-artifact@v4
        with:
          name: claude-security-results
          path: claude-security-results.json

  upload-claude-findings:
    runs-on: ubuntu-latest
    needs: [skylos, claude-security]
    steps:
      - name: Checkout
        uses: actions/checkout@v4

      - name: Setup Python
        uses: actions/setup-python@v5
        with:
          python-version: '{python_version}'

      - name: Install Skylos
        run: pip install skylos

      - name: Download Claude Security Results
        uses: actions/download-artifact@v4
        with:
          name: claude-security-results

      - name: Download Skylos Results
        uses: actions/download-artifact@v4
        with:
          name: skylos-results

      - name: Ingest Claude Security Findings
        run: skylos ingest claude-security --input claude-security-results.json --cross-reference skylos-results.json
        env:
          SKYLOS_TOKEN: ${{{{ secrets.SKYLOS_TOKEN }}}}
"""


def _build_trigger_block(triggers: list[str]) -> str:
    lines = ['"on":']
    for trigger in triggers:
        if trigger == "pull_request":
            lines.append("  pull_request:")
            lines.append("    branches: [main]")
        elif trigger == "push":
            lines.append("  push:")
            lines.append("    branches: [main]")
        elif trigger == "schedule":
            lines.append("  schedule:")
            lines.append("    - cron: '0 6 * * 1'  # Weekly Monday 6am")
        elif trigger == "workflow_dispatch":
            lines.append("  workflow_dispatch:")
        else:
            lines.append(f"  {trigger}:")
    return "\n".join(lines)


def write_workflow(content: str, output_path: str):
    path = Path(output_path)
    path.parent.mkdir(parents=True, exist_ok=True)

    if path.exists():
        console.print(f"[yellow]Overwriting existing workflow: {path}[/yellow]")

    path.write_text(content)
    console.print(f"[bold green]Workflow written to {path}[/bold green]")
    console.print(f"[dim]Commit and push to activate: git add {path} && git push[/dim]")



import os
import subprocess
import sys

from rich.console import Console
from rich.prompt import Confirm, Prompt

try:
    import inquirer

    INTERACTIVE = True
except ImportError:
    INTERACTIVE = False

console = Console()


def run_cmd(cmd_list, error_msg="Git command failed"):
    try:
        result = subprocess.run(cmd_list, check=True, capture_output=True, text=True)
        return result.stdout.strip()
    except subprocess.CalledProcessError as e:
        console.print(f"[bold red]Error:[/bold red] {error_msg}\n[dim]{e.stderr}[/dim]")
        return None


def get_git_status():
    out = run_cmd(
        ["git", "status", "--porcelain"], "Could not get git status. Is this a repo?"
    )
    if not out:
        return []

    files = []
    for line in out.splitlines():
        if len(line) > 3:
            files.append(line[3:])
    return files


def run_push():
    console.print("[dim]Pushing to remote...[/dim]")
    try:
        subprocess.run(["git", "push"], check=True)
        console.print("[bold green] Deployment Complete. Code is live.[/bold green]")
    except subprocess.CalledProcessError:
        console.print(
            "[bold red] Push failed. Check your git remote settings.[/bold red]"
        )


def start_deployment_wizard():
    if not INTERACTIVE:
        console.print(
            "[yellow]Install 'inquirer' (pip install inquirer) to use interactive deployment.[/yellow]"
        )
        return

    console.print("\n[bold cyan] Skylos Deployment Wizard[/bold cyan]")

    files = get_git_status()
    if not files:
        console.print("[green]Working tree is clean.[/green]")
        if Confirm.ask("Push existing commits?"):
            run_push()
        return

    q_scope = [
        inquirer.List(
            "scope",
            message="What do you want to stage?",
            choices=[
                "All changed files",
                "Select files manually",
                "Skip commit (Push only)",
            ],
        ),
    ]
    ans_scope = inquirer.prompt(q_scope)
    if not ans_scope:
        return

    if ans_scope["scope"] == "Select files manually":
        q_files = [inquirer.Checkbox("files", message="Select files", choices=files)]
        ans_files = inquirer.prompt(q_files)
        if not ans_files or not ans_files["files"]:
            console.print("[red]No files selected.[/red]")
            return
        run_cmd(["git", "add"] + ans_files["files"])
        console.print(f"[green]Staged {len(ans_files['files'])} files.[/green]")

    elif ans_scope["scope"] == "All changed files":
        run_cmd(["git", "add", "."])
        console.print("[green]Staged all files.[/green]")

    if ans_scope["scope"] != "Skip commit (Push only)":
        msg = Prompt.ask("[bold green]Enter commit message[/bold green]")
        if not msg:
            console.print("[red]Commit message required.[/red]")
            return
        if run_cmd(["git", "commit", "-m", msg]):
            console.print("[green]✓ Committed.[/green]")

    if Confirm.ask("Ready to git push?"):
        run_push()


def _get_finding_file(finding):
    return finding.get("file", finding.get("file_path", ""))


def _check_agent_gate(findings_lists, agent_file_set, agent_cfg, reasons):
    """Evaluate agent-specific thresholds against findings in AI-authored files.

    Returns False if any agent threshold is exceeded.
    """
    agent_passed = True

    agent_danger = []
    agent_quality = []
    agent_secrets = []
    agent_dead_code = 0

    for category, items in findings_lists.items():
        for f in items:
            fpath = _get_finding_file(f)
            if fpath not in agent_file_set:
                continue
            if category == "danger":
                agent_danger.append(f)
            elif category == "quality":
                agent_quality.append(f)
            elif category == "secrets":
                agent_secrets.append(f)
            elif category == "dead_code":
                agent_dead_code += 1

    agent_critical = [
        d for d in agent_danger if str(d.get("severity", "")).lower() == "critical"
    ]
    agent_high = [
        d for d in agent_danger if str(d.get("severity", "")).lower() == "high"
    ]

    a_max_critical = agent_cfg.get("max_critical")
    if isinstance(a_max_critical, int) and len(agent_critical) > a_max_critical:
        agent_passed = False
        reasons.append(
            f"Agent gate: {len(agent_critical)} critical issue(s) in AI-authored files (max: {a_max_critical})"
        )

    a_max_high = agent_cfg.get("max_high")
    if isinstance(a_max_high, int) and len(agent_high) > a_max_high:
        agent_passed = False
        reasons.append(
            f"Agent gate: {len(agent_high)} high severity issue(s) in AI-authored files (max: {a_max_high})"
        )

    a_max_security = agent_cfg.get("max_security")
    if isinstance(a_max_security, int) and len(agent_danger) > a_max_security:
        agent_passed = False
        reasons.append(
            f"Agent gate: {len(agent_danger)} security issue(s) in AI-authored files (max: {a_max_security})"
        )

    a_max_quality = agent_cfg.get("max_quality")
    if isinstance(a_max_quality, int) and len(agent_quality) > a_max_quality:
        agent_passed = False
        reasons.append(
            f"Agent gate: {len(agent_quality)} quality issue(s) in AI-authored files (max: {a_max_quality})"
        )

    a_max_secrets = agent_cfg.get("max_secrets")
    if isinstance(a_max_secrets, int) and len(agent_secrets) > a_max_secrets:
        agent_passed = False
        reasons.append(
            f"Agent gate: {len(agent_secrets)} secret(s) in AI-authored files (max: {a_max_secrets})"
        )

    a_max_dead_code = agent_cfg.get("max_dead_code")
    if isinstance(a_max_dead_code, int) and agent_dead_code > a_max_dead_code:
        agent_passed = False
        reasons.append(
            f"Agent gate: {agent_dead_code} dead code issue(s) in AI-authored files (max: {a_max_dead_code})"
        )

    min_defend = agent_cfg.get("min_defend_score")
    require_defend = agent_cfg.get("require_defend", False)
    if require_defend or isinstance(min_defend, (int, float)):
        reasons.append(
            "Agent gate: require_defend/min_defend_score set but no defense data available (run skylos defend)"
        )
        agent_passed = False

    return agent_passed


def check_gate(results, config, strict=False, provenance=None):
    results = results or {}
    config = config or {}

    reasons = []
    passed = True

    total_findings = sum(
        len(results.get(k, []))
        for k in (
            "unused_functions",
            "unused_imports",
            "unused_variables",
            "unused_classes",
            "unused_parameters",
        )
    )

    danger = results.get("danger", []) or []
    quality = results.get("quality", []) or []
    secrets = results.get("secrets", []) or []

    critical_issues = []
    high_issues = []

    for issue in danger:
        sev = str(issue.get("severity", "")).lower()
        if sev == "critical":
            critical_issues.append(issue)
        elif sev == "high":
            high_issues.append(issue)

    gate_config = config.get("gate", {}) if config else {}

    if strict:
        total_issues = total_findings + len(danger) + len(quality) + len(secrets)
        if total_issues > 0:
            return False, [f"Strict mode: {total_issues} issue(s) found"]
        return True, []

    fail_on_critical = gate_config.get("fail_on_critical", True)
    max_critical = gate_config.get("max_critical", 0)
    max_high = gate_config.get("max_high", 5)
    max_security = gate_config.get("max_security", 10)
    max_quality = gate_config.get("max_quality", 10)
    max_secrets = gate_config.get("max_secrets", None)
    max_dead_code = gate_config.get("max_dead_code", None)

    if fail_on_critical and len(critical_issues) > 0:
        passed = False
        reasons.append(f"{len(critical_issues)} critical security issue(s)")

    elif isinstance(max_critical, int) and len(critical_issues) > max_critical:
        passed = False
        reasons.append(f"{len(critical_issues)} critical issues (max: {max_critical})")

    if isinstance(max_high, int) and len(high_issues) > max_high:
        passed = False
        reasons.append(f"{len(high_issues)} high severity issues (max: {max_high})")

    if isinstance(max_security, int) and len(danger) > max_security:
        passed = False
        reasons.append(f"{len(danger)} total security issues (max: {max_security})")

    if isinstance(max_quality, int) and len(quality) > max_quality:
        passed = False
        reasons.append(f"{len(quality)} quality issues (max: {max_quality})")

    if isinstance(max_secrets, int) and len(secrets) > max_secrets:
        passed = False
        reasons.append(f"{len(secrets)} secrets issues (max: {max_secrets})")

    if isinstance(max_dead_code, int) and total_findings > max_dead_code:
        passed = False
        reasons.append(f"{total_findings} dead code issue(s) (max: {max_dead_code})")

    # Agent-aware gating: apply stricter thresholds to AI-authored files
    agent_cfg = gate_config.get("agent")
    if provenance and agent_cfg and provenance.agent_files:
        agent_file_set = set(provenance.agent_files)

        dead_code_items = []
        for k in (
            "unused_functions",
            "unused_imports",
            "unused_variables",
            "unused_classes",
            "unused_parameters",
        ):
            dead_code_items.extend(results.get(k, []) or [])

        findings_lists = {
            "danger": danger,
            "quality": quality,
            "secrets": secrets,
            "dead_code": dead_code_items,
        }

        agent_passed = _check_agent_gate(
            findings_lists, agent_file_set, agent_cfg, reasons
        )
        if not agent_passed:
            passed = False

    return passed, reasons


def build_summary_markdown(results, passed, reasons):
    results = results or {}

    danger = results.get("danger", []) or []
    quality = results.get("quality", []) or []
    secrets = results.get("secrets", []) or []

    critical_count = sum(
        1 for d in danger if str(d.get("severity", "")).lower() == "critical"
    )
    high_count = sum(1 for d in danger if str(d.get("severity", "")).lower() == "high")
    dead_code_count = sum(
        len(results.get(k, []) or [])
        for k in (
            "unused_functions",
            "unused_imports",
            "unused_classes",
            "unused_variables",
            "unused_parameters",
        )
    )

    status = "PASSED" if passed else "FAILED"
    icon = "✅" if passed else "❌"

    lines = [
        "## Skylos Analysis Results",
        "",
        "| Category | Count | Status |",
        "|----------|-------|--------|",
        f"| Security (critical) | {critical_count} | {'✅' if critical_count == 0 else '❌'} |",
        f"| Security (high) | {high_count} | {'✅' if high_count <= 5 else '⚠️'} |",
        f"| Security (total) | {len(danger)} | {'✅' if len(danger) <= 10 else '⚠️'} |",
        f"| Quality | {len(quality)} | {'✅' if len(quality) <= 10 else '⚠️'} |",
        f"| Secrets | {len(secrets)} | {'✅' if len(secrets) == 0 else '❌'} |",
        f"| Dead Code | {dead_code_count} | ℹ️ |",
        "",
        f"**Result: {icon} {status}**",
    ]

    if reasons:
        lines.append("")
        lines.append("### Failure Reasons")
        for r in reasons:
            lines.append(f"- {r}")

    return "\n".join(lines)


def write_github_summary(markdown):
    summary_path = os.environ.get("GITHUB_STEP_SUMMARY")
    if summary_path:
        try:
            with open(summary_path, "a") as f:
                f.write(markdown + "\n")
        except OSError as e:
            console.print(
                f"[yellow]Could not write to GITHUB_STEP_SUMMARY: {e}[/yellow]"
            )
    else:
        console.print(markdown)


def run_gate_interaction(
    *,
    results=None,
    result=None,
    config=None,
    strict=False,
    force=False,
    command_to_run=None,
    summary=False,
    provenance=None,
):
    console = Console()

    if results is None:
        results = result or {}

    config = config or {}
    gate_cfg = config.get("gate") or {}

    strict = bool(strict or gate_cfg.get("strict", False))

    try:
        passed, reasons = check_gate(
            results, config, strict=strict, provenance=provenance
        )
    except TypeError:
        passed, reasons = check_gate(results, config)

    if summary:
        md = build_summary_markdown(results, passed, reasons)
        write_github_summary(md)

    if passed:
        console.print("\n[bold green]✅ Quality Gate: PASSED[/bold green]")

        if command_to_run:
            proc = subprocess.run(command_to_run)
            return getattr(proc, "returncode", 0)

        return 0

    console.print("\n[bold red] Quality Gate: FAILED[/bold red]")
    for r in reasons or []:
        console.print(f"   • {r}")

    if force:
        console.print("[yellow] Forced pass (local only)[/yellow]")
        return 0

    if strict:
        return 1

    try:
        if sys.stdout.isatty():
            if Confirm.ask("Quality gate failed. Continue anyway?"):
                start_deployment_wizard()
                return 0
            return 1
    except Exception:
        pass

    return 1



from __future__ import annotations

import json
import logging
import time
from dataclasses import dataclass, field, asdict
from pathlib import Path
from typing import Any

logger = logging.getLogger(__name__)

TRIAGE_DIR = ".skylos"
TRIAGE_FILE = "triage_patterns.json"

MIN_OBSERVATIONS_SUGGEST = 3
MIN_OBSERVATIONS_AUTO = 5
MIN_CONFIDENCE_SUGGEST = 0.7
MIN_CONFIDENCE_AUTO = 0.85


@dataclass
class TriagePattern:
    pattern_type: str
    pattern: str
    rule_id: str
    action: str
    confidence: float = 0.0
    observations: int = 0
    last_updated: float = 0.0

    def to_dict(self) -> dict[str, Any]:
        return asdict(self)

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> TriagePattern:
        return cls(
            pattern_type=data.get("pattern_type", ""),
            pattern=data.get("pattern", ""),
            rule_id=data.get("rule_id", ""),
            action=data.get("action", ""),
            confidence=data.get("confidence", 0.0),
            observations=data.get("observations", 0),
            last_updated=data.get("last_updated", 0.0),
        )


def _pattern_key(pattern: TriagePattern) -> str:
    return (
        f"{pattern.pattern_type}:{pattern.pattern}:{pattern.rule_id}:{pattern.action}"
    )


class TriageLearner:
    def __init__(self) -> None:
        self._patterns: dict[str, TriagePattern] = {}

    @property
    def pattern_count(self) -> int:
        return len(self._patterns)

    def load(self, project_root: str | Path) -> None:
        path = Path(project_root) / TRIAGE_DIR / TRIAGE_FILE
        if not path.exists():
            return
        try:
            data = json.loads(path.read_text(encoding="utf-8"))
            for entry in data.get("patterns", []):
                pattern = TriagePattern.from_dict(entry)
                key = _pattern_key(pattern)
                self._patterns[key] = pattern
            logger.debug("Loaded %d triage patterns", len(self._patterns))
        except Exception as e:
            logger.debug("Failed to load triage patterns: %s", e)

    def save(self, project_root: str | Path) -> None:
        """Save patterns to disk."""
        path = Path(project_root) / TRIAGE_DIR / TRIAGE_FILE
        path.parent.mkdir(parents=True, exist_ok=True)
        data = {
            "version": 1,
            "patterns": [p.to_dict() for p in self._patterns.values()],
        }
        try:
            path.write_text(json.dumps(data, indent=2), encoding="utf-8")
        except Exception as e:
            logger.debug("Failed to save triage patterns: %s", e)

    def learn_from_triage(
        self,
        finding: dict[str, Any],
        action: str,
    ) -> list[TriagePattern]:
        if action not in ("dismiss", "accept"):
            return []

        updated: list[TriagePattern] = []
        candidates = self._extract_patterns(finding, action)

        for candidate in candidates:
            key = _pattern_key(candidate)
            existing = self._patterns.get(key)

            if existing is not None:
                existing.observations += 1
                existing.last_updated = time.time()
                existing.confidence = existing.observations / (
                    existing.observations + 1
                )
                updated.append(existing)
            else:
                candidate.observations = 1
                candidate.confidence = 0.5
                candidate.last_updated = time.time()
                self._patterns[key] = candidate
                updated.append(candidate)

        return updated

    def _extract_patterns(
        self,
        finding: dict[str, Any],
        action: str,
    ) -> list[TriagePattern]:
        patterns: list[TriagePattern] = []
        rule_id = finding.get("rule_id", "")
        file_path = finding.get("file", "")
        name = finding.get("simple_name", finding.get("name", ""))

        if not rule_id:
            return patterns

        if file_path:
            parts = file_path.replace("\\", "/").split("/")
            if len(parts) > 1:
                dir_pattern = parts[0] + "/**"
                patterns.append(
                    TriagePattern(
                        pattern_type="file_glob",
                        pattern=dir_pattern,
                        rule_id=rule_id,
                        action=action,
                    )
                )

            if "." in file_path:
                ext = "." + file_path.rsplit(".", 1)[-1]
                if any(
                    marker in file_path
                    for marker in ("test_", "_test.", "tests/", "test/")
                ):
                    patterns.append(
                        TriagePattern(
                            pattern_type="file_glob",
                            pattern=f"**/test_*{ext}",
                            rule_id=rule_id,
                            action=action,
                        )
                    )

        if name:
            if name.startswith("test_"):
                patterns.append(
                    TriagePattern(
                        pattern_type="name_pattern",
                        pattern="test_*",
                        rule_id=rule_id,
                        action=action,
                    )
                )
            elif name.startswith("__") and name.endswith("__"):
                patterns.append(
                    TriagePattern(
                        pattern_type="name_pattern",
                        pattern="__*__",
                        rule_id=rule_id,
                        action=action,
                    )
                )
            elif name.startswith("_"):
                patterns.append(
                    TriagePattern(
                        pattern_type="name_pattern",
                        pattern="_*",
                        rule_id=rule_id,
                        action=action,
                    )
                )

        decorators = finding.get("decorators", [])
        if isinstance(decorators, list):
            for dec in decorators:
                dec_str = str(dec).strip()
                if dec_str:
                    patterns.append(
                        TriagePattern(
                            pattern_type="decorator",
                            pattern=dec_str,
                            rule_id=rule_id,
                            action=action,
                        )
                    )

        return patterns

    def predict_triage(
        self,
        finding: dict[str, Any],
    ) -> tuple[str, float] | None:
        best_match: tuple[str, float] | None = None
        best_confidence = 0.0

        for pattern in self._patterns.values():
            if pattern.observations < MIN_OBSERVATIONS_SUGGEST:
                continue
            if pattern.confidence < MIN_CONFIDENCE_SUGGEST:
                continue

            if self._matches(finding, pattern):
                if pattern.confidence > best_confidence:
                    best_confidence = pattern.confidence
                    best_match = (pattern.action, pattern.confidence)

        return best_match

    def get_auto_triage_candidates(
        self,
        findings: list[dict[str, Any]],
    ) -> list[tuple[dict[str, Any], str, float]]:
        candidates: list[tuple[dict[str, Any], str, float]] = []

        for finding in findings:
            prediction = self.predict_triage(finding)
            if prediction is None:
                continue
            action, confidence = prediction
            if confidence >= MIN_CONFIDENCE_AUTO:
                matching_patterns = [
                    p
                    for p in self._patterns.values()
                    if (
                        p.observations >= MIN_OBSERVATIONS_AUTO
                        and p.confidence >= MIN_CONFIDENCE_AUTO
                        and self._matches(finding, p)
                    )
                ]
                if matching_patterns:
                    candidates.append((finding, action, confidence))

        return candidates

    def _matches(self, finding: dict[str, Any], pattern: TriagePattern) -> bool:
        rule_id = finding.get("rule_id", "")
        if pattern.rule_id and pattern.rule_id != rule_id:
            return False

        if pattern.pattern_type == "file_glob":
            file_path = finding.get("file", "")
            return _glob_match(file_path, pattern.pattern)

        elif pattern.pattern_type == "name_pattern":
            name = finding.get("simple_name", finding.get("name", ""))
            return _glob_match(name, pattern.pattern)

        elif pattern.pattern_type == "decorator":
            decorators = finding.get("decorators", [])
            if isinstance(decorators, list):
                return any(str(d).strip() == pattern.pattern for d in decorators)
            return False

        elif pattern.pattern_type == "rule_type":
            return True  # rule_id already matched above

        return False

    def get_patterns(self) -> list[TriagePattern]:
        return list(self._patterns.values())


def _glob_match(text: str, pattern: str) -> bool:
    import fnmatch

    return fnmatch.fnmatch(text, pattern)



import shutil
import subprocess

import pytest

from skylos.agent_center import (
    clear_action_triage,
    build_headline,
    build_ranked_actions,
    detect_changed_files,
    load_agent_state,
    rebuild_agent_state_from_existing,
    save_agent_state,
    snapshot_file_signatures,
    update_action_triage,
)


def test_detect_changed_files_includes_new_changed_and_removed_files():
    previous = {
        "src/a.py": {"mtime_ns": 1, "size": 10},
        "src/b.py": {"mtime_ns": 2, "size": 20},
    }
    current = {
        "src/a.py": {"mtime_ns": 1, "size": 10},
        "src/c.py": {"mtime_ns": 3, "size": 30},
        "src/b.py": {"mtime_ns": 9, "size": 20},
    }

    changed = detect_changed_files(previous, current)

    assert changed == ["src/b.py", "src/c.py"]


def test_snapshot_file_signatures_skips_gitignored_files(tmp_path):
    if shutil.which("git") is None:
        pytest.skip("git is required for this test")

    project = tmp_path / "repo"
    project.mkdir()
    ignored_dir = project / "customenv"
    kept_dir = project / "src"
    ignored_dir.mkdir(parents=True)
    kept_dir.mkdir(parents=True)
    (project / ".gitignore").write_text("customenv/\n", encoding="utf-8")
    (ignored_dir / "ghost.py").write_text("x = 1\n", encoding="utf-8")
    (kept_dir / "keep.py").write_text("x = 1\n", encoding="utf-8")
    subprocess.run(["git", "init", "-q"], cwd=project, check=True)

    signatures = snapshot_file_signatures(project)

    assert "src/keep.py" in signatures
    assert "customenv/ghost.py" not in signatures


def test_build_ranked_actions_prioritizes_new_critical_changed_security():
    findings = [
        {
            "fingerprint": "security:1",
            "rule_id": "SKY-D999",
            "category": "security",
            "severity": "CRITICAL",
            "message": "SQL injection path",
            "file": "src/auth.py",
            "absolute_file": "/tmp/src/auth.py",
            "line": 14,
            "confidence": 95,
            "is_new_vs_baseline": True,
            "is_new_since_last_scan": True,
            "is_in_changed_file": True,
        },
        {
            "fingerprint": "dead:1",
            "rule_id": "SKY-U002",
            "category": "dead_code",
            "severity": "INFO",
            "message": "Unused import: os",
            "file": "src/helpers.py",
            "absolute_file": "/tmp/src/helpers.py",
            "line": 2,
            "confidence": 40,
            "is_new_vs_baseline": False,
            "is_new_since_last_scan": False,
            "is_in_changed_file": False,
        },
    ]

    actions = build_ranked_actions(findings, changed_files=["src/auth.py"])

    assert actions[0]["file"] == "src/auth.py"
    assert actions[0]["severity"] == "CRITICAL"
    assert actions[0]["action_type"] == "inspect_now"
    assert actions[0]["score"] > actions[1]["score"]


def test_build_headline_prefers_urgent_changed_code():
    headline = build_headline(
        critical=1,
        high=1,
        new_total=2,
        changed_total=2,
        baseline_present=True,
        total=10,
    )

    assert headline == "2 urgent finding(s) need attention in changed code"


def test_agent_state_round_trip(tmp_path):
    project_root = tmp_path / "repo"
    project_root.mkdir()
    state = {
        "summary": {"headline": "1 urgent finding"},
        "actions": [{"id": "a1", "title": "Review HIGH SKY-D201"}],
    }

    save_agent_state(project_root, state)
    loaded = load_agent_state(project_root)

    assert loaded == state


def test_rebuild_agent_state_filters_dismissed_and_snoozed_actions():
    state = {
        "project_root": "/tmp/repo",
        "file_signatures": {"src/auth.py": {"mtime_ns": 1, "size": 10}},
        "changed_files": ["src/auth.py"],
        "baseline_present": True,
        "findings": [
            {
                "fingerprint": "security:1",
                "rule_id": "SKY-D999",
                "category": "security",
                "severity": "CRITICAL",
                "message": "SQL injection path",
                "file": "src/auth.py",
                "absolute_file": "/tmp/repo/src/auth.py",
                "line": 14,
                "confidence": 95,
                "is_new_vs_baseline": True,
                "is_new_since_last_scan": True,
                "is_in_changed_file": True,
            },
            {
                "fingerprint": "dead:1",
                "rule_id": "SKY-U002",
                "category": "dead_code",
                "severity": "INFO",
                "message": "Unused import: os",
                "file": "src/helpers.py",
                "absolute_file": "/tmp/repo/src/helpers.py",
                "line": 2,
                "confidence": 40,
                "is_new_vs_baseline": False,
                "is_new_since_last_scan": False,
                "is_in_changed_file": False,
            },
        ],
        "triage": {
            "security:1": {
                "status": "dismissed",
                "updated_at": "2026-03-16T00:00:00+00:00",
            },
            "dead:1": {
                "status": "snoozed",
                "updated_at": "2026-03-16T00:00:00+00:00",
                "snoozed_until": "2099-03-16T00:00:00+00:00",
            },
        },
    }

    rebuilt = rebuild_agent_state_from_existing(state)

    assert rebuilt["actions"] == []
    assert rebuilt["summary"]["dismissed"] == 1
    assert rebuilt["summary"]["snoozed"] == 1


def test_update_and_clear_action_triage_round_trip(tmp_path):
    project_root = tmp_path / "repo"
    project_root.mkdir()
    state = {
        "project_root": str(project_root),
        "file_signatures": {"src/auth.py": {"mtime_ns": 1, "size": 10}},
        "changed_files": ["src/auth.py"],
        "baseline_present": False,
        "findings": [
            {
                "fingerprint": "security:1",
                "rule_id": "SKY-D999",
                "category": "security",
                "severity": "HIGH",
                "message": "Shell injection path",
                "file": "src/auth.py",
                "absolute_file": str(project_root / "src" / "auth.py"),
                "line": 9,
                "confidence": 91,
                "is_new_vs_baseline": True,
                "is_new_since_last_scan": True,
                "is_in_changed_file": True,
            },
        ],
    }
    save_agent_state(project_root, rebuild_agent_state_from_existing(state))

    dismissed = update_action_triage(project_root, "security:1", status="dismissed")
    assert dismissed["actions"] == []
    assert dismissed["triage"]["security:1"]["status"] == "dismissed"

    restored = clear_action_triage(project_root, "security:1")
    assert restored["actions"][0]["id"] == "security:1"
    assert restored["triage"] == {}



"""Defense plugins — each checks for a specific missing guardrail."""

from skylos.defend.plugins.no_dangerous_sink import NoDangerousSinkPlugin
from skylos.defend.plugins.tool_scope import ToolScopePlugin
from skylos.defend.plugins.tool_schema_present import ToolSchemaPresentPlugin
from skylos.defend.plugins.prompt_delimiter import PromptDelimiterPlugin
from skylos.defend.plugins.output_validation import OutputValidationPlugin
from skylos.defend.plugins.model_pinned import ModelPinnedPlugin
from skylos.defend.plugins.input_length_limit import InputLengthLimitPlugin

# Phase 3: extended defense plugins
from skylos.defend.plugins.untrusted_input_to_prompt import UntrustedInputToPromptPlugin
from skylos.defend.plugins.rag_context_isolation import RagContextIsolationPlugin
from skylos.defend.plugins.output_pii_filter import OutputPiiFilterPlugin

# Phase 3: ops plugins
from skylos.defend.plugins.logging_present import LoggingPresentPlugin
from skylos.defend.plugins.cost_controls import CostControlsPlugin
from skylos.defend.plugins.rate_limiting import RateLimitingPlugin

ALL_PLUGINS = [
    # Defense plugins
    NoDangerousSinkPlugin(),
    ToolScopePlugin(),
    ToolSchemaPresentPlugin(),
    PromptDelimiterPlugin(),
    OutputValidationPlugin(),
    ModelPinnedPlugin(),
    InputLengthLimitPlugin(),
    UntrustedInputToPromptPlugin(),
    RagContextIsolationPlugin(),
    OutputPiiFilterPlugin(),
    # Ops plugins
    LoggingPresentPlugin(),
    CostControlsPlugin(),
    RateLimitingPlugin(),
]

__all__ = ["ALL_PLUGINS"]



from skylos.defend.plugin import DefensePlugin
from skylos.defend.result import DefenseResult
from skylos.discover.integration import LLMIntegration
from skylos.discover.graph import AIIntegrationGraph


class ToolScopePlugin(DefensePlugin):
    id = "tool-scope"
    name = "Tool Scope Restricted"
    severity = "critical"
    owasp_llm = "LLM04"
    description = "Tool functions must not have unrestricted access to shell, filesystem, or network"
    remediation = (
        "Wrap dangerous operations in tool functions with permission checks, "
        "allowlists, or sandboxing. Restrict tool capabilities to minimum required scope."
    )

    def applies_to(self, integration: LLMIntegration) -> bool:
        return integration.integration_type == "agent" and len(integration.tools) > 0

    def check(
        self, integration: LLMIntegration, graph: AIIntegrationGraph
    ) -> DefenseResult:
        dangerous_tools = []
        for tool in integration.tools:
            if tool.dangerous_calls:
                calls = ", ".join(tool.dangerous_calls)
                dangerous_tools.append(f"{tool.name}: {calls}")

        if not dangerous_tools:
            return self._pass(
                integration,
                integration.location,
                "No dangerous operations in tool functions",
            )

        details = "; ".join(dangerous_tools)
        return self._fail(
            integration,
            integration.location,
            f"Tools with unrestricted dangerous operations: {details}",
        )



from __future__ import annotations

import sys
from pathlib import Path

from .planner import RemediationPlanner, RemediationPlan, FixBatch
from .executor import RemediationExecutor
from .prompts import build_pr_description


class RemediationAgent:
    def __init__(
        self,
        *,
        model: str = "gpt-4.1",
        api_key: str | None = None,
        test_cmd: str | None = None,
        severity_filter: str | None = None,
        provider: str | None = None,
        base_url: str | None = None,
    ):
        self.model = model
        self.api_key = api_key
        self.test_cmd = test_cmd
        self.severity_filter = severity_filter
        self.provider = provider
        self.base_url = base_url
        self.planner = RemediationPlanner(severity_filter=severity_filter)

    def run(
        self,
        path: str | Path,
        *,
        dry_run: bool = False,
        max_fixes: int = 10,
        auto_pr: bool = False,
        branch_prefix: str = "skylos/fix",
        quiet: bool = False,
    ) -> dict:
        path = Path(path).resolve()
        log = _logger(quiet)

        log("Step 1/5: Scanning project...")
        results = self._scan(path)
        danger_count = len(results.get("danger", []) or [])
        quality_count = len(results.get("quality", []) or [])
        secrets_count = len(results.get("secrets", []) or [])
        total = danger_count + quality_count + secrets_count
        log(
            f"  Found {total} findings ({danger_count} danger, "
            f"{quality_count} quality, {secrets_count} secrets)"
        )

        if total == 0:
            log("No findings — nothing to remediate.")
            return RemediationPlan().summary()

        log("Step 2/5: Creating remediation plan...")
        plan = self.planner.create_plan(results, max_fixes=max_fixes)
        log(
            f"  Planned {sum(len(b.findings) for b in plan.batches)} fixes "
            f"across {len(plan.batches)} files "
            f"({plan.skipped_findings} skipped)"
        )

        if dry_run:
            log("\n[Dry run] No changes applied. Plan:")
            self._print_plan(plan, log)
            return plan.summary()

        branch_name = None
        if auto_pr:
            log("Step 3/6: Creating remediation branch...")
            branch_name = self._prepare_pr_branch(
                path if path.is_dir() else path.parent,
                branch_prefix,
                log,
            )
            if not branch_name:
                summary = plan.summary()
                summary["branch_error"] = "Could not create remediation branch"
                return summary

        log("Step 4/6: Generating and applying fixes...")
        executor = RemediationExecutor(
            test_cmd=self.test_cmd,
            project_root=path if path.is_dir() else path.parent,
        )

        fixer = self._create_fixer()
        fixed_files: list[str] = []

        for batch in plan.batches:
            log(f"\n  Fixing {batch.file} ({len(batch.findings)} findings)...")
            self._process_batch(batch, fixer, executor, log)

            if batch.status == "fixed":
                fixed_files.append(batch.file)
                log(f"    ✓ Fixed successfully")
            else:
                log(f"    ✗ Status: {batch.status} — {batch.fix_description}")

        fixed_count = sum(1 for b in plan.batches if b.status == "fixed")
        log(f"\n  Results: {fixed_count}/{len(plan.batches)} files fixed")

        pr_url = None
        if auto_pr and fixed_files:
            log("Step 5/6: Creating pull request...")
            pr_url = self._create_pr(executor, plan, branch_name, log)
        else:
            if not fixed_files:
                log("Step 5/6: No fixes applied — skipping PR.")
            else:
                log("Step 5/6: Skipped (--auto-pr not set).")

        log("Step 6/6: Summary")
        summary = plan.summary()
        if pr_url:
            summary["pr_url"] = pr_url
        if branch_name:
            summary["branch"] = branch_name
        self._print_plan(plan, log)

        return summary

    def _scan(self, path: Path) -> dict:
        import json
        from skylos.analyzer import analyze as run_analyze

        raw = run_analyze(
            str(path),
            conf=0,
            enable_danger=True,
            enable_quality=True,
            enable_secrets=True,
        )
        return json.loads(raw) if isinstance(raw, str) else raw

    def _create_fixer(self):
        from .agents import FixerAgent, AgentConfig

        config = AgentConfig()
        config.model = self.model
        if self.api_key:
            config.api_key = self.api_key
        if self.provider:
            config.provider = self.provider
        if self.base_url:
            config.base_url = self.base_url
        return FixerAgent(config)

    def _process_batch(
        self,
        batch: FixBatch,
        fixer,
        executor: RemediationExecutor,
        log,
    ):
        file_path = batch.file
        p = Path(file_path)
        if not p.exists():
            batch.status = "failed"
            batch.fix_description = "File not found"
            return

        source = p.read_text(encoding="utf-8")
        batch.source = source

        primary = batch.findings[0]

        try:
            fix = fixer.fix(
                source,
                file_path,
                primary.line,
                primary.message,
            )
        except Exception as e:
            batch.status = "failed"
            batch.fix_description = f"Fix generation error: {e}"
            return

        if fix is None:
            batch.status = "skipped"
            batch.fix_description = "LLM could not generate a fix"
            return

        confidence = getattr(fix.confidence, "value", str(fix.confidence))
        if confidence == "low":
            batch.status = "skipped"
            batch.fix_description = "Fix confidence too low"
            return

        if not executor.apply_fix(file_path, fix.fixed_code):
            batch.status = "failed"
            batch.fix_description = "Could not write fixed file"
            return

        test_result = executor.run_tests()
        if not test_result.passed:
            executor.revert_fix(file_path)
            batch.status = "test_failed"
            batch.fix_description = f"Tests failed after fix ({test_result.command})"
            return

        rule_ids = [f.rule_id for f in batch.findings]
        verify = executor.verify_fix(file_path, rule_ids)
        if not verify.finding_resolved:
            executor.revert_fix(file_path)
            batch.status = "not_resolved"
            batch.fix_description = (
                f"Finding still present after fix: {verify.remaining_rule_ids}"
            )
            return

        batch.status = "fixed"
        desc = getattr(fix, "description", "")
        batch.fix_description = desc[:200] if desc else primary.message

    def _prepare_pr_branch(
        self,
        project_root: Path,
        branch_prefix: str,
        log,
    ) -> str | None:
        executor = RemediationExecutor(project_root=project_root)
        try:
            branch = executor.create_branch(branch_prefix)
            log(f"  Branch: {branch}")
            return branch
        except Exception as e:
            log(f"  Failed to create branch: {e}")
            return None

    def _create_pr(
        self,
        executor: RemediationExecutor,
        plan: RemediationPlan,
        branch: str | None,
        log,
    ) -> str | None:
        fixed_batches = [b for b in plan.batches if b.status == "fixed"]
        if not fixed_batches or not branch:
            return None

        files = [b.file for b in fixed_batches]
        finding_count = sum(len(b.findings) for b in fixed_batches)

        commit_msg = (
            f"fix: remediate {finding_count} security/quality findings\n\n"
            f"Automated by Skylos DevOps Agent.\n"
            f"Files: {', '.join(Path(f).name for f in files)}"
        )

        if not executor.commit_fixes(commit_msg, files):
            log("  Failed to commit.")
            return None

        if not executor.push_branch(branch):
            log("  Failed to push branch.")
            return None

        summary = plan.summary()
        body = build_pr_description(summary)
        title = f"fix: skylos remediation ({finding_count} issues)"
        pr_url = executor.create_pr(branch, title, body)
        if pr_url:
            log(f"  PR created: {pr_url}")
        return pr_url

    def _print_plan(self, plan: RemediationPlan, log):
        s = plan.summary()
        log(f"\n  Total findings: {s['total_findings']}")
        log(f"  Planned: {s['planned']}")
        log(f"  Fixed: {s['fixed']}")
        log(f"  Failed: {s['failed']}")
        log(f"  Skipped: {s['skipped']}")

        if s["batches"]:
            log("")
            log(f"  {'File':<40} {'Status':<15} {'Sev':<10} {'#':<5}")
            log(f"  {'─' * 40} {'─' * 15} {'─' * 10} {'─' * 5}")
            for b in s["batches"]:
                fname = b["file"]
                if len(fname) > 38:
                    fname = "…" + fname[-37:]
                log(
                    f"  {fname:<40} {b['status']:<15} "
                    f"{b['top_severity']:<10} {b['findings']:<5}"
                )


def _logger(quiet: bool):
    if quiet:
        return lambda msg: None

    def _log(msg):
        print(msg, file=sys.stderr)

    return _log



from __future__ import annotations
import ast
import sys


def _qualified_name(node):
    func = node.func
    parts = []
    while isinstance(func, ast.Attribute):
        parts.append(func.attr)
        func = func.value
    if isinstance(func, ast.Name):
        parts.append(func.id)
        parts.reverse()
        return ".".join(parts)
    return None


class _JWTChecker(ast.NodeVisitor):
    def __init__(self, file_path, findings):
        self.file_path = file_path
        self.findings = findings

    def generic_visit(self, node):
        for field, value in ast.iter_fields(node):
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, ast.AST):
                        self.visit(item)
            elif isinstance(value, ast.AST):
                self.visit(value)

    def _report(self, node, message, severity="HIGH"):
        self.findings.append(
            {
                "rule_id": "SKY-D232",
                "severity": severity,
                "message": message,
                "file": str(self.file_path),
                "line": node.lineno,
                "col": node.col_offset,
            }
        )

    def visit_Call(self, node):
        qn = _qualified_name(node)
        if not qn or not qn.endswith((".decode", ".encode")):
            self.generic_visit(node)
            return

        if not qn.startswith("jwt.") and not qn.endswith("jwt.decode"):
            self.generic_visit(node)
            return

        for kw in node.keywords:
            if kw.arg == "algorithms" and isinstance(kw.value, ast.List):
                for elt in kw.value.elts:
                    if isinstance(elt, ast.Constant) and isinstance(elt.value, str):
                        if elt.value.lower() == "none":
                            self._report(
                                node,
                                'JWT vulnerability: algorithms=["none"] allows unsigned tokens.',
                                severity="CRITICAL",
                            )

            if kw.arg == "verify" and isinstance(kw.value, ast.Constant):
                if kw.value.value is False:
                    self._report(
                        node,
                        "JWT vulnerability: verify=False disables signature verification.",
                        severity="CRITICAL",
                    )

            if kw.arg == "options" and isinstance(kw.value, ast.Dict):
                for key, val in zip(kw.value.keys, kw.value.values):
                    if (
                        isinstance(key, ast.Constant)
                        and key.value == "verify_signature"
                        and isinstance(val, ast.Constant)
                        and val.value is False
                    ):
                        self._report(
                            node,
                            "JWT vulnerability: verify_signature=False disables signature verification.",
                            severity="CRITICAL",
                        )

        self.generic_visit(node)


def scan(tree, file_path, findings):
    try:
        checker = _JWTChecker(file_path, findings)
        checker.visit(tree)
    except Exception as e:
        print(f"JWT analysis failed for {file_path}: {e}", file=sys.stderr)



from pathlib import Path

from skylos.grader import (
    score_to_letter,
    compute_grade,
    score_security,
    score_quality,
    score_dead_code,
    score_dependencies,
    score_secrets,
    count_lines_of_code,
    _interpolate_dead_code_score,
    generate_badge_url,
    generate_badge_svg,
    CATEGORY_WEIGHTS,
)


def _empty_result(**overrides):
    base = {
        "danger": [],
        "quality": [],
        "secrets": [],
        "dependency_vulnerabilities": [],
        "unused_functions": [],
        "unused_imports": [],
        "unused_classes": [],
        "unused_variables": [],
        "unused_parameters": [],
    }
    base.update(overrides)
    return base


class TestScoreToLetter:
    def test_a_plus(self):
        assert score_to_letter(100) == "A+"
        assert score_to_letter(97) == "A+"

    def test_a(self):
        assert score_to_letter(93) == "A"
        assert score_to_letter(96) == "A"

    def test_a_minus(self):
        assert score_to_letter(90) == "A-"
        assert score_to_letter(92) == "A-"

    def test_b_plus(self):
        assert score_to_letter(87) == "B+"
        assert score_to_letter(89) == "B+"

    def test_b(self):
        assert score_to_letter(83) == "B"
        assert score_to_letter(86) == "B"

    def test_b_minus(self):
        assert score_to_letter(80) == "B-"
        assert score_to_letter(82) == "B-"

    def test_c_range(self):
        assert score_to_letter(77) == "C+"
        assert score_to_letter(73) == "C"
        assert score_to_letter(70) == "C-"

    def test_d_range(self):
        assert score_to_letter(67) == "D+"
        assert score_to_letter(63) == "D"
        assert score_to_letter(60) == "D-"

    def test_f(self):
        assert score_to_letter(59) == "F"
        assert score_to_letter(30) == "F"
        assert score_to_letter(0) == "F"

    def test_clamp_over_100(self):
        assert score_to_letter(105) == "A+"

    def test_clamp_negative(self):
        assert score_to_letter(-5) == "F"


class TestScoreSecurity:
    def test_no_findings(self):
        score, issue = score_security([])
        assert score == 100

    def test_critical_caps_at_55(self):
        findings = [{"severity": "CRITICAL", "message": "SQL injection"}]
        score, _ = score_security(findings)
        assert score <= 55

    def test_high_caps_at_79(self):
        findings = [{"severity": "HIGH", "message": "Weak hash"}]
        score, _ = score_security(findings)
        assert score <= 79

    def test_multiple_criticals_stack(self):
        findings = [
            {"severity": "CRITICAL", "message": "SQL injection"},
            {"severity": "CRITICAL", "message": "Command injection"},
        ]
        score, _ = score_security(findings)
        # cap 55, then -15 -15 = 25
        assert score == 25

    def test_mixed_severities(self):
        findings = [
            {"severity": "HIGH", "message": "XSS"},
            {"severity": "MEDIUM", "message": "Open redirect"},
            {"severity": "LOW", "message": "Info disclosure"},
        ]
        score, _ = score_security(findings)
        # cap 79, then -8 -3 -1 = 67
        assert score == 67

    def test_key_issue_is_worst(self):
        findings = [
            {"severity": "LOW", "message": "Info disclosure"},
            {"severity": "CRITICAL", "message": "RCE found"},
        ]
        _, issue = score_security(findings)
        assert "RCE" in issue

    def test_floor_at_zero(self):
        findings = [{"severity": "CRITICAL", "message": f"vuln{i}"} for i in range(10)]
        score, _ = score_security(findings)
        assert score == 0


class TestScoreQuality:
    def test_no_findings(self):
        score, _ = score_quality([], 10000)
        assert score == 100

    def test_normalized_per_1k_loc(self):
        findings = [{"severity": "MEDIUM", "message": "Complex"}]
        score_small, _ = score_quality(findings, 500)
        score_large, _ = score_quality(findings, 50000)
        assert score_large > score_small

    def test_critical_quality_caps(self):
        findings = [{"severity": "CRITICAL", "message": "Extreme complexity"}]
        score, _ = score_quality(findings, 10000)
        assert score <= 55


class TestScoreDeadCode:
    def test_no_dead_code(self):
        result = _empty_result()
        score, _ = score_dead_code(result, 10000)
        assert score == 100

    def test_moderate_density(self):
        result = _empty_result(unused_functions=[{"name": f"f{i}"} for i in range(10)])
        score, _ = score_dead_code(result, 2000)
        assert score == 85

    def test_high_density_low_score(self):
        result = _empty_result(unused_functions=[{"name": f"f{i}"} for i in range(50)])
        score, _ = score_dead_code(result, 1000)
        assert score == 0

    def test_key_issue_format(self):
        result = _empty_result(unused_functions=[{"name": f"f{i}"} for i in range(5)])
        _, issue = score_dead_code(result, 5000)
        assert "5 dead symbols" in issue
        assert "/1K LOC" in issue

    def test_zero_loc_uses_raw_count(self):
        result = _empty_result(unused_functions=[{"name": f"f{i}"} for i in range(3)])
        score, _ = score_dead_code(result, 0)
        assert score < 100


class TestInterpolation:
    def test_zero_density(self):
        assert _interpolate_dead_code_score(0) == 100

    def test_exact_breakpoints(self):
        assert _interpolate_dead_code_score(5) == 85
        assert _interpolate_dead_code_score(15) == 55
        assert _interpolate_dead_code_score(30) == 20
        assert _interpolate_dead_code_score(50) == 0

    def test_midpoint(self):
        score = _interpolate_dead_code_score(10.0)
        assert 55 < score < 85

    def test_beyond_max(self):
        assert _interpolate_dead_code_score(100) == 0

    def test_negative(self):
        assert _interpolate_dead_code_score(-5) == 100


class TestScoreDependencies:
    def test_no_findings(self):
        score, _ = score_dependencies([])
        assert score == 100

    def test_critical_vuln(self):
        findings = [{"severity": "CRITICAL", "message": "CVE-2024-1234"}]
        score, _ = score_dependencies(findings)
        assert score <= 55


class TestScoreSecrets:
    def test_no_secrets(self):
        score, _ = score_secrets([])
        assert score == 100

    def test_one_secret_caps_at_69(self):
        score, _ = score_secrets([{"message": "AWS key"}])
        assert score == 69

    def test_additional_subtract_20(self):
        secrets = [{"message": f"Secret {i}"} for i in range(3)]
        score, _ = score_secrets(secrets)
        # 69 - 20 - 20 = 29
        assert score == 29

    def test_floor_at_zero(self):
        secrets = [{"message": f"Secret {i}"} for i in range(10)]
        score, _ = score_secrets(secrets)
        assert score == 0

    def test_key_issue_single(self):
        _, issue = score_secrets([{"message": "AWS key found"}])
        assert "AWS key" in issue

    def test_key_issue_multiple(self):
        secrets = [{"message": "s1"}, {"message": "s2"}]
        _, issue = score_secrets(secrets)
        assert "2 secrets" in issue


class TestComputeGrade:
    def test_clean_codebase_a_plus(self):
        result = _empty_result()
        grade = compute_grade(result, 10000)
        assert grade["overall"]["letter"] == "A+"
        assert grade["overall"]["score"] == 100

    def test_critical_security_caps_overall(self):
        result = _empty_result(danger=[{"severity": "CRITICAL", "message": "RCE"}])
        grade = compute_grade(result, 10000)
        assert grade["overall"]["score"] <= 79
        assert grade["overall"]["letter"][0] in ("C", "D", "F")

    def test_weights_sum_to_one(self):
        assert abs(sum(CATEGORY_WEIGHTS.values()) - 1.0) < 0.001

    def test_all_categories_present(self):
        grade = compute_grade(_empty_result(), 1000)
        for cat in ("security", "quality", "dead_code", "dependencies", "secrets"):
            assert cat in grade["categories"]
            cat_data = grade["categories"][cat]
            assert "score" in cat_data
            assert "letter" in cat_data
            assert "weight" in cat_data
            assert "key_issue" in cat_data

    def test_total_loc_included(self):
        grade = compute_grade(_empty_result(), 42000)
        assert grade["total_loc"] == 42000

    def test_secrets_subgrade_capped(self):
        result = _empty_result(secrets=[{"message": "API key"}])
        grade = compute_grade(result, 10000)
        assert grade["categories"]["secrets"]["score"] <= 69

    def test_mixed_issues_weighted(self):
        result = _empty_result(
            danger=[{"severity": "HIGH", "message": "XSS"}],
            quality=[{"severity": "MEDIUM", "message": "Complex"}],
            unused_functions=[{"name": f"f{i}"} for i in range(10)],
        )
        grade = compute_grade(result, 5000)
        assert 0 < grade["overall"]["score"] < 100
        assert grade["categories"]["security"]["score"] < 100
        assert grade["categories"]["quality"]["score"] < 100
        assert grade["categories"]["dead_code"]["score"] < 100


class TestCountLinesOfCode:
    def test_python_file(self, tmp_path):
        f = tmp_path / "test.py"
        f.write_text("x = 1\n\n# comment\ny = 2\n")
        assert count_lines_of_code([f]) == 2

    def test_typescript_file(self, tmp_path):
        f = tmp_path / "test.ts"
        f.write_text("const x = 1;\n\n// comment\nconst y = 2;\n")
        assert count_lines_of_code([f]) == 2

    def test_empty_file(self, tmp_path):
        f = tmp_path / "empty.py"
        f.write_text("")
        assert count_lines_of_code([f]) == 0

    def test_nonexistent_file(self):
        assert count_lines_of_code([Path("/nonexistent/file.py")]) == 0

    def test_multiple_files(self, tmp_path):
        f1 = tmp_path / "a.py"
        f1.write_text("x = 1\ny = 2\n")
        f2 = tmp_path / "b.py"
        f2.write_text("z = 3\n")
        assert count_lines_of_code([f1, f2]) == 3


class TestBadge:
    def test_url_format(self):
        url = generate_badge_url("A+", 98)
        assert "img.shields.io" in url
        assert "brightgreen" in url

    def test_url_f_grade(self):
        url = generate_badge_url("F", 30)
        assert "red" in url

    def test_svg_contains_grade(self):
        svg = generate_badge_svg("B+", 84)
        assert "Skylos" in svg
        assert "B+" in svg
        assert "<svg" in svg
